import itertools
from typing import AbstractSet, Callable, Tuple

import numpy as np
import pandas as pd
import pytest

from ...exceptions import (
    NegativeFormulaNotNamedRelationException,
    NegativeFormulaNotSafeRangeException,
    NeuroLangException,
    UnsupportedProgramError,
    UnsupportedQueryError
)
from ...logic.horn_clauses import Fol2DatalogTranslationException
from ...probabilistic import dalvi_suciu_lift
from ...probabilistic.exceptions import (
    ForbiddenConditionalQueryNonConjunctive,
    RepeatedTuplesInProbabilisticRelationError
)
from ...regions import SphericalVolume
from ...utils.relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet,
    RelationalAlgebraFrozenSet
)
from ..probabilistic_frontend import NeurolangPDL


def assert_almost_equal(set_a, set_b):
    for tupl_a in set_a:
        assert any(
            all(
                np.isclose(term_a, term_b)
                if isinstance(term_a, (float, np.float32, np.float64))
                else term_a == term_b
                for term_a, term_b in zip(tupl_a, tupl_b)
            )
            for tupl_b in set_b
        )


def test_add_uniform_probabilistic_choice_set():
    nl = NeurolangPDL()

    prob = [(a,) for a in range(10)]
    prob_set = nl.add_uniform_probabilistic_choice_over_set(prob, "prob")
    res = nl[prob_set]

    assert prob_set.type is AbstractSet[Tuple[float, int]]
    assert res.type is AbstractSet[Tuple[float, int]]
    assert res.value == frozenset((1 / len(prob), a) for a in range(10))

    d = [("a",), ("b",), ("c",), ("d",)]
    data = nl.add_uniform_probabilistic_choice_over_set(d, name="data")
    res_d = nl[data]

    assert data.type is AbstractSet[Tuple[float, str]]
    assert res_d.type is AbstractSet[Tuple[float, str]]

    d = pd.DataFrame([("a",), ("b",), ("c",), ("d",)], columns=["ids"])
    data2 = nl.add_uniform_probabilistic_choice_over_set(d, name="data2")
    res_d2 = nl[data]

    assert data2.type is AbstractSet[Tuple[float, str]]
    assert res_d2.type is AbstractSet[Tuple[float, str]]


def test_deterministic_query():
    nl = NeurolangPDL()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_tuple_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_tuple_set(d2, name="data2")

    with nl.scope as e:
        e.query1[e.y] = data1[e.x] & data2[e.x, e.y]
        res = nl.solve_all()

    assert "query1" in res.keys()
    q1 = res["query1"].as_pandas_dataframe().values
    assert len(q1) == 4
    for elem in q1:
        assert elem[0] in ["a", "b", "c", "d"]


def test_probabilistic_query():
    nl = NeurolangPDL()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_uniform_probabilistic_choice_over_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_uniform_probabilistic_choice_over_set(d2, name="data2")

    d3 = [("a",), ("b",), ("c",)]
    data3 = nl.add_uniform_probabilistic_choice_over_set(d3, name="data3")

    with nl.scope as e:
        e.query1[e.y, e.PROB[e.y]] = data1[e.x] & data2[e.x, e.y]
        e.query2[e.y, e.PROB[e.y]] = e.query1[e.y] & data3[e.y]
        with pytest.raises(UnsupportedProgramError):
            nl.solve_all()


def test_marg_query():
    nl = NeurolangPDL()
    nl.add_probabilistic_choice_from_tuples(
        {(0.2, "a"), (0.3, "b"), (0.5, "c")}, name="P"
    )
    nl.add_tuple_set({(4, "a"), (5, "a"), (5, "b")}, name="Q")
    nl.add_probabilistic_facts_from_tuples(
        {
            (0.2, 1, "a"),
            (0.1, 1, "b"),
            (0.7, 2, "b"),
            (0.9, 2, "c"),
        },
        name="R",
    )

    with nl.scope as e:
        e.Z[e.x, e.z, e.PROB[e.x, e.z]] = (
            e.Q(e.x, e.y) & e.P(e.y)
        ) // (e.R(e.z, e.y) & e.P(e.y))
        res = nl.solve_all()

    expected = RelationalAlgebraFrozenSet(
        {
            (4, 1, 0.2 * 0.2 / (0.2 * 0.2 + 0.3 * 0.1)),
            (5, 1, (0.2 * 0.2 + 0.3 * 0.1) / (0.2 * 0.2 + 0.3 * 0.1)),
            (5, 2, 0.3 * 0.7 / (0.3 * 0.7 + 0.9 * 0.5)),
        }
    )
    assert_almost_equal(res["Z"], expected)


def test_mixed_queries():
    nl = NeurolangPDL()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_tuple_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_tuple_set(d2, name="data2")

    d3 = [("a",), ("b",), ("c",), ("d",)]
    data3 = nl.add_uniform_probabilistic_choice_over_set(d3, name="data3")

    with nl.scope as e:
        e.tmp[e.x, e.y] = data1[e.x] & data2[e.x, e.y]
        e.query1[e.x, e.y, e.PROB[e.x, e.y]] = e.tmp[e.x, e.y]
        e.query2[e.y, e.PROB[e.y]] = e.tmp[e.x, e.y] & data3[e.y]
        res = nl.solve_all()

    assert "query1" in res.keys()
    assert "query2" in res.keys()
    assert len(res["query2"].columns) == 2
    q2 = res["query2"].as_pandas_dataframe().values
    assert len(q2) == 4
    for elem in q2:
        assert elem[1] == 0.25
        assert elem[0] in ["a", "b", "c", "d"]


def test_simple_within_language_succ_query():
    nl = NeurolangPDL()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("e",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        res = nl.solve_all()
    assert "Z" in res.keys()
    df = res["Z"].as_pandas_dataframe()
    assert len(df) == 1
    assert tuple(df.values[0])[0] == "a"
    assert np.isclose(tuple(df.values[0])[1], 1 / 9)


def test_within_language_succ_query():
    nl = NeurolangPDL()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [
            ("a", "b"),
            ("b", "c"),
            ("b", "d"),
        ],
        name="P",
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x, e.y] & Q[e.x]
        res = nl.query((e.x, e.p), e.Z(e.x, e.p))
    expected = RelationalAlgebraFrozenSet(
        [
            ("b", 1 / 3 * 1 / 2 + 1 / 3 * 1 / 2),
            ("a", 1 / 3 * 1 / 2),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_query():
    nl = NeurolangPDL()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        res = nl.query((e.x, e.p), e.Z[e.x, e.p])
        res_2 = nl.query(e.p, e.Z(e.x, e.p))
    expected = RelationalAlgebraFrozenSet(
        [
            ("a", 1 / 9),
            ("c", 1 / 9),
        ]
    )
    assert_almost_equal(res, expected)
    assert res.row_type == Tuple[str, float]
    assert_almost_equal(res_2, expected.projection(1))
    assert res_2.row_type == Tuple[float]


def test_solve_query_prob_col_not_last():
    nl = NeurolangPDL()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.PROB[e.x], e.x] = P[e.x] & Q[e.x]
        res = nl.query((e.p, e.x), e.Z[e.p, e.x])
    expected = RelationalAlgebraFrozenSet(
        [
            (1 / 9, "a"),
            (1 / 9, "c"),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_boolean_query():
    nl = NeurolangPDL()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.ans[e.PROB()] = P[e.x] & Q[e.x]
        res = nl.query((e.p), e.ans[e.p])

    expected = RelationalAlgebraFrozenSet([(2 / 9,)])
    assert_almost_equal(res, expected)


def test_solve_query_with_constant():
    nl = NeurolangPDL()
    P = nl.add_probabilistic_choice_from_tuples(
        {(0.2, "a"), (0.3, "b"), (0.5, "c")}, name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.ans[e.x, e.PROB(e.x)] = P[e.x] & Q[e.x]
        res = nl.query((e.p), e.ans["a", e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            (0.2 * 1/3,),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_complex_stratified_query():
    """
    R(1, 2) : 0.3
    R(1, 4) : 0.7
    R(2, 2) : 0.2
    R(2, 4) : 0.6
    R(2, 6) : 0.8
    Q(4) : 0.2 v Q(6) : 0.8
    A(x, PROB[x])       :- Q(x), R(1, x), R(2, x)
    B(x, y, PROB[x, y]) :- Q(y), R(1, x), R(2, x)
    C(x, y, p1, p2)     :- A(x, p1), B(x, y, p2)

    """
    nl = NeurolangPDL()
    R = nl.add_probabilistic_facts_from_tuples(
        [(0.3, 1, 2), (0.7, 1, 4), (0.2, 2, 2), (0.6, 2, 4), (0.8, 2, 6)],
        name="R",
    )
    Q = nl.add_probabilistic_choice_from_tuples([(0.2, 4), (0.8, 6)], name="Q")
    with nl.scope as e:
        e.A[e.x, e.PROB[e.x]] = Q[e.x] & R[1, e.x] & R[2, e.x]
        e.B[e.x, e.y, e.PROB[e.x, e.y]] = Q[e.y] & R[1, e.x] & R[2, e.x]
        e.C[e.x, e.y, e.p1, e.p2] = e.A[e.x, e.p1] & e.B[e.x, e.y, e.p2]
        res = nl.query((e.x, e.y, e.h, e.z), e.C[e.x, e.y, e.h, e.z])
    expected = RelationalAlgebraFrozenSet(
        [
            (4, 4, 0.2 * 0.7 * 0.6, 0.2 * 0.7 * 0.6),
            (4, 6, 0.2 * 0.7 * 0.6, 0.8 * 0.7 * 0.6),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_complex_stratified_query_with_deterministic_part():
    nl = NeurolangPDL()
    A = nl.add_tuple_set([("a",), ("b",), ("c",)], name="A")
    B = nl.add_tuple_set([("a",), ("b",)], name="B")
    P = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.8, "b")],
        name="P",
    )
    with nl.scope as e:
        e.C[e.x, e.y] = A[e.x] & A[e.y] & B[e.x]
        e.D[e.x, e.y, e.PROB[e.x, e.y]] = e.C[e.x, e.y] & P[e.y]
        res = nl.query((e.x, e.y, e.p), e.D[e.x, e.y, e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            ("a", "a", 0.2),
            ("a", "b", 0.8),
            ("b", "a", 0.2),
            ("b", "b", 0.8),
        ]
    )
    assert_almost_equal(res, expected)


def test_neurolange_dl_deterministic_negation():
    neurolang = NeurolangPDL()
    s = neurolang.new_symbol(name="s")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")

    dataset = {(i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset, name="q")
    s[x, y] = ~q(x, x) & q(x, y)

    res = neurolang.solve_all()

    assert res["s"].to_unnamed() == {(i, j) for i, j in dataset if i != j}


def test_neurolange_dl_probabilistic_negation():
    neurolang = NeurolangPDL()
    s = neurolang.new_symbol(name="s")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")
    prob = neurolang.new_symbol(name="PROB")

    dataset_det = {(i, i * 2) for i in range(10)}
    dataset = {((1 + i) / 10, i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset_det, name="q")
    r = neurolang.add_probabilistic_facts_from_tuples(dataset, name="r")

    s[x, y, prob(x, y)] = ~r(x, y) & q(x, y)

    result = neurolang.solve_all()["s"].to_unnamed()
    expected = {(i, j, 1 - p) for (p, i, j) in dataset}
    assert_almost_equal(result, expected)


@pytest.mark.skip
def test_neurolange_dl_probabilistic_negation_not_safe():
    neurolang = NeurolangPDL()
    s = neurolang.new_symbol(name="s")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")
    z = neurolang.new_symbol(name="z")
    prob = neurolang.new_symbol(name="PROB")

    dataset_det = {(i, i * 2) for i in range(10)}
    dataset = {((1 + i) / 10, i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset_det, name="q")
    r = neurolang.add_probabilistic_facts_from_tuples(dataset, name="r")

    s[x, y, prob(x, y)] = ~r(x, z) & q(x, y)

    with pytest.raises(NegativeFormulaNotSafeRangeException):
        neurolang.solve_all()


@pytest.mark.skip
def test_neurolange_dl_probabilistic_negation_rule():
    neurolang = NeurolangPDL()
    s = neurolang.new_symbol(name="s")
    t = neurolang.new_symbol(name="t")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")
    z = neurolang.new_symbol(name="z")
    prob = neurolang.new_symbol(name="PROB")

    dataset_det = {(i, i * 2) for i in range(10)}
    dataset = {((1 + i) / 10, i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset_det, name="q")
    r = neurolang.add_probabilistic_facts_from_tuples(dataset, name="r")

    t[x, y] = r(x, y) & q(y, z)
    s[x, y, prob(x, y)] = ~t(x, x) & q(x, y)

    with pytest.raises(NegativeFormulaNotNamedRelationException):
        neurolang.solve_all()


def test_neurolang_dl_aggregation():
    neurolang = NeurolangPDL()
    q = neurolang.new_symbol(name="q")
    p = neurolang.new_symbol(name="p")
    r = neurolang.new_symbol(name="r")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")

    @neurolang.add_symbol
    def sum_(x):
        return sum(x)

    for i in range(10):
        q[i % 2, i] = True

    p[x, sum_(y)] = q[x, y]

    sol = neurolang.query(r(x, y), p(x, y))

    res_q = {(0, 2 + 4 + 8), (1, 1 + 5 + 9)}

    assert len(sol) == 2
    assert sol[r] == res_q
    assert sol[p] == res_q


def test_post_probabilistic_aggregation():
    nl = NeurolangPDL()
    A = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.9, "b"), (0.5, "c")],
        name="A",
    )
    B = nl.add_probabilistic_choice_from_tuples(
        [(0.2, "a", "c"), (0.7, "b", "c"), (0.1, "a", "d")],
        name="B",
    )

    @nl.add_symbol
    def mysum(x):
        return sum(x)

    with nl.scope as e:
        e.C[e.x, e.y, e.PROB[e.x, e.y]] = A[e.x] & B[e.x, e.y]
        e.D[e.x, e.mysum(e.p)] = e.C[e.x, e.y, e.p]
        res = nl.query((e.x, e.s), e.D[e.x, e.s])

    assert len(res) == 2
    expected = {("a", 0.2 * 0.2 + 0.2 * 0.1), ("b", 0.9 * 0.7)}
    assert_almost_equal(res.to_unnamed(), expected)


def test_empty_result_query():
    nl = NeurolangPDL()
    A = nl.add_tuple_set([("f",), ("d",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "a"),
            (0.7, "b"),
            (0.6, "c"),
        ],
        name="B",
    )
    with nl.scope as e:
        e.Q[e.x, e.PROB[e.x]] = A[e.x] & B[e.x]
        res = nl.query((e.x, e.p), e.Q[e.x, e.p])
    assert res.is_empty()
    with nl.scope as e:
        e.Q[e.x] = A[e.x] & A["b"]
        res = nl.query((e.x,), e.Q[e.x])
    assert res.is_empty()


def test_forbidden_query_on_probabilistic_predicate():
    nl = NeurolangPDL()
    A = nl.add_tuple_set([("f",), ("d",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.7, "b"), (0.6, "c")], name="B"
    )
    with pytest.raises(UnsupportedQueryError):
        with nl.scope as e:
            e.Q[e.x] = A[e.x] & B[e.x] & A["b"]
            nl.query((e.x,), e.Q[e.x])


def test_empty_boolean_query_result():
    nl = NeurolangPDL()
    A = nl.add_tuple_set([("a",), ("b",), ("c",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [(0.4, "a"), (0.5, "b")], name="B"
    )
    C = nl.add_probabilistic_choice_from_tuples(
        [(0.8, "c"), (0.2, "f")], name="C"
    )
    with nl.scope as e:
        e.D[e.x, e.PROB[e.x]] = A[e.x] & B[e.x] & C[e.x]
        res = nl.query(tuple(), e.D[e.x, e.p])
    assert not res


def test_trivial_probability_query_result():
    nl = NeurolangPDL()
    b = set(((0.4, "a"), (0.5, "b")))
    B = nl.add_probabilistic_facts_from_tuples(b, name='B')
    with nl.scope as e:
        e.D[e.x, e.PROB[e.x]] = B[e.x]
        res = nl.query((e.p, e.x), e.D[e.x, e.p])
    assert set(res) == b


def test_equality():
    nl = NeurolangPDL()
    r1 = nl.add_tuple_set([(i,) for i in range(5)], name="r1")

    with nl.scope as e:
        e.r2[e.x] = r1(e.y) & (e.x == e.y * 2)

        sol = nl.query((e.x,), e.r2[e.x])

    assert set(sol) == set((2 * i,) for i in range(5))

    r2 = nl.add_tuple_set([("Hola",), ("Hello",), ("Bonjour",)], name="r2")

    @nl.add_symbol
    def lower(input: str) -> str:
        return input.lower()

    with nl.scope as e:
        e.r3[e.x] = r2(e.y) & (e.x == lower(e.y))

        sol = nl.query((e.x,), e.r3[e.x])

    assert set(sol) == set((("hola",), ("hello",), ("bonjour",)))


def test_equality2():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [("Hola", "var"), ("Hello", "var2"), ("Bonjour", "var")],
        name="test_var",
    )

    @nl.add_symbol
    def word_lower(word: str) -> str:
        return str(word).lower()

    with nl.scope as e:
        e.low[e.lower_] = e.test_var[e.word, "var"] & (
            e.lower_ == word_lower(e.word)
        )

        query = nl.query((e.lower_,), e.low[e.lower_])

    assert set(query) == set((("hola",), ("bonjour",)))


def test_result_both_deterministic_and_post_probabilistic():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            ("this", "is", "inglese"),
            ("questo", "è", "italiano"),
            ("ceci", "est", "francese"),
        ],
        name="parole",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.4, "valentino"),
            (0.6, "pietro"),
        ],
        name="persona_scelta",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.1, "italiano"),
            (0.7, "inglese"),
            (0.2, "francese"),
        ],
        name="lingua_parlata",
    )
    with nl.scope as e:
        e.e_una_lingua[e.lingua] = e.parole[e.subject, e.verb, e.lingua]
        e.lingua_da_lua_decisa[
            e.persona, e.lingua, e.PROB[e.persona, e.lingua]
        ] = (
            e.e_una_lingua[e.lingua]
            & e.persona_scelta[e.persona]
            & e.lingua_parlata[e.lingua]
        )
        e.utilizzare_le_probabilita[e.lingua, e.p] = e.lingua_da_lua_decisa[
            ..., e.lingua, e.p
        ] & (e.p > 0.1)
        res = nl.solve_all()
    assert "e_una_lingua" in res
    assert "lingua_da_lua_decisa" in res
    assert "utilizzare_le_probabilita" in res
    assert len(res["utilizzare_le_probabilita"]) == 3
    expected = RelationalAlgebraFrozenSet(
        [
            ("francese", 0.12),
            ("inglese", 0.7 * 0.4),
            ("inglese", 0.7 * 0.6),
        ]
    )
    assert_almost_equal(res["utilizzare_le_probabilita"], expected)


def test_result_query_relation_correct_column_names():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with nl.environment as e:
        e.climbs[e.p, e.PROB[e.p, e.city], e.city] = (
            e.person[e.p] & e.lives_in[e.p, e.city] & e.does_not_smoke[e.p]
        )
        solution = nl.solve_all()
    assert all(name in solution for name in ["climbs", "person", "lives_in"])


def test_add_constraints_and_rewrite():
    nl = NeurolangPDL()

    nl.add_tuple_set(
        [
            ("neurolang"),
        ],
        name="project",
    )

    nl.add_tuple_set(
        [
            ("neurolang", "db"),
            ("neurolang", "logic"),
        ],
        name="inArea",
    )

    with nl.scope as e:
        nl.add_constraint(
            e.project[e.x] & e.inArea[e.x, e.y],
            e.hasCollaborator[e.z, e.y, e.x],
        )

        e.p[e.b] = e.hasCollaborator[e.a, "db", e.b]
        res = nl.solve_all()

    assert list(res["p"])[0] == ("neurolang",)


def test_solve_marg_query():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.2, "alice", "running"),
            (0.8, "bob", "climbing"),
        ],
        name="practice",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with nl.environment as e:
        e.query[e.p, e.PROB[e.p, e.city, e.sport], e.city, e.sport] = (
            e.person[e.p] & e.lives_in[e.p, e.city] & e.does_not_smoke[e.p]
        ) // e.practice[e.p, e.sport]
        solution = nl.solve_all()
    assert all(name in solution for name in ["query", "person", "lives_in"])
    with nl.environment as e:
        result = nl.query(
            (e.p, e.prob, e.city, e.sport),
            e.query(e.p, e.prob, e.city, e.sport),
        )
    expected = RelationalAlgebraFrozenSet(
        iterable=[
            ("alice", 0.9, "paris", "running"),
            ("bob", 0.8, "marseille", "climbing"),
        ],
    )
    assert_almost_equal(result, expected)


def test_query_based_pfact():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            (2, 0.2),
            (7, 0.8),
            (4, 0.4),
        ],
        name="A",
    )
    with nl.environment as e:
        (e.B @ (e.p / 2))[e.x] = e.A[e.x, e.p]
        e.Query[e.PROB[e.x], e.x] = e.B[e.x]
        result = nl.query((e.x, e.p), e.Query[e.p, e.x])
    expected = RelationalAlgebraFrozenSet(
        [
            (2, 0.1),
            (7, 0.4),
            (4, 0.2),
        ]
    )
    assert_almost_equal(result, expected)


def test_query_based_pfact_empty():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            (2, 0.2),
            (7, 0.8),
            (4, 0.4),
        ],
        name="A",
    )
    with nl.environment as e:
        (e.B @ (e.p / 2))[e.x] = e.A[e.x, e.p] & (e.p > 0.8)
        e.Query[e.PROB[e.x], e.x] = e.B[e.x]
        result = nl.query((e.x, e.p), e.Query[e.p, e.x])
    expected = RelationalAlgebraFrozenSet(
        []
    )
    assert_almost_equal(result, expected)


def test_query_based_pfact_region_volume():
    nl = NeurolangPDL()

    @nl.add_symbol
    def volume(s: SphericalVolume) -> float:
        return (4 / 3) * np.pi * s.radius ** 3

    assert volume.symbol_name in nl.functions

    nl.add_tuple_set(
        [
            ("contained", SphericalVolume((0, 0, 0), 1)),
            ("container", SphericalVolume((0, 0, 0), 2)),
        ],
        name="my_sphere",
    )

    with nl.environment as e:
        (e.Z @ (e.volume[e.contained] / e.volume[e.container]))[
            e.contained, e.container
        ] = (
            e.my_sphere["contained", e.contained]
            & e.my_sphere["container", e.container]
        )
        e.Query[
            e.contained, e.container, e.PROB[e.contained, e.container]
        ] = e.Z[e.contained, e.container]
        res = nl.query((e.p,), e.Query[e.contained, e.container, e.p])
    expected = NamedRelationalAlgebraFrozenSet(("p",), [(1 / 2 ** 3,)])
    assert_almost_equal(res, expected)


def test_solve_marg_query_disjunction():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.2, "alice", "running"),
            (0.8, "bob", "climbing"),
        ],
        name="practice",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with pytest.raises(
        (
            ForbiddenConditionalQueryNonConjunctive,
            Fol2DatalogTranslationException,
        )
    ):
        with nl.environment as e:
            e.query[e.p, e.PROB[e.p, e.city, e.sport], e.city, e.sport] = (
                e.person[e.p]
                & (e.lives_in[e.p, e.city] | e.does_not_smoke[e.p])
            ) // e.practice[e.p, e.sport]


def test_query_without_safe_plan():
    nl = NeurolangPDL()
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "alice"),
            (0.8, "bob"),
        ],
        name="names",
    )

    expected = NamedRelationalAlgebraFrozenSet(
        columns=('x', 'y', 'PROB'),
        iterable={
            ('alice', 'alice', 0.2),
            ('bob', 'bob', 0.8),
            ('alice', 'bob', 0.8 * 0.2),
            ('bob', 'alice', 0.8 * 0.2)
        }
    )

    with nl.environment as e:
        e.q[e.x, e.y, e.PROB[e.x, e.y]] = e.names[e.x] & e.names[e.y]

    res = nl.solve_all()['q']
    assert res == expected


def test_query_without_safe_plan_2():
    nl = NeurolangPDL()
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "alice", 1),
            (0.4, "alice", 2),
            (0.8, "bob", 2),
            (0.6, "bob", 1),
        ],
        name="names",
    )

    expected = NamedRelationalAlgebraFrozenSet(
        columns=('x', 'y', 'PROB'),
        iterable={
            ('alice', 'alice', 0.52),
            ('bob', 'bob', 0.92),
            ('alice', 'bob', 0.40160000000000007),
            ('bob', 'alice', 0.40160000000000007)
        }
    )

    with nl.environment as e:
        e.q[e.x, e.y, e.PROB[e.x, e.y]] = e.names[e.x, e.s] & e.names[e.y, e.s]

    res = nl.solve_all()['q']
    assert res == expected


def test_query_without_safe_fails():
    nl = NeurolangPDL(
        probabilistic_solvers=(dalvi_suciu_lift.solve_succ_query,),
        probabilistic_marg_solvers=(dalvi_suciu_lift.solve_marg_query,),
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "alice"),
            (0.8, "bob"),
        ],
        name="names",
    )

    with nl.environment as e:
        e.q[e.x, e.y, e.PROB[e.x, e.y]] = e.names[e.x] & e.names[e.y]

    with pytest.raises(dalvi_suciu_lift.NonLiftableException):
        nl.solve_all()


def test_query_self_join_matrix_query_ds():
    nl = NeurolangPDL(
        probabilistic_solvers=(dalvi_suciu_lift.solve_succ_query,),
        probabilistic_marg_solvers=(dalvi_suciu_lift.solve_marg_query,),
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "alice", 1),
            (0.8, "bob", 1),
            (0.4, "sam", 2),
            (0.3, "alice", 1)
        ],
        name="names",
    )

    with nl.environment as e:
        e.q[e.PROB()] = e.names[e.x, e.z] & e.names[e.y, e.z]

    res = nl.solve_all()['q']

    assert len(res) == 1
    assert res.arity == 1
    assert list(res)[0].PROB == 0.9328


def test_cbma_two_term_conjunctive_query():
    nl = NeurolangPDL()
    nl.add_uniform_probabilistic_choice_over_set(
        [
            ("s1",),
            ("s2",),
            ("s3",),
        ],
        name="SelectedStudy",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.1, "t1", "s1"),
            (0.2, "t2", "s1"),
            (0.3, "t1", "s2"),
            (0.4, "t2", "s2"),
            (0.5, "t1", "s3"),
            (0.6, "t2", "s3"),
        ],
        name="TermInStudy",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.6, "v1", "s1"),
            (0.5, "v2", "s1"),
            (0.4, "v1", "s2"),
            (0.3, "v2", "s2"),
            (0.2, "v1", "s3"),
            (0.1, "v2", "s3"),
        ],
        name="VoxelReported",
    )
    with nl.environment as e:
        e.TermAssociation[e.t] = e.SelectedStudy[e.s] & e.TermInStudy[e.t, e.s]
        e.Activation[e.v] = e.SelectedStudy[e.s] & e.VoxelReported[e.v, e.s]
        e.probmap[e.v, e.PROB[e.v]] = (e.Activation[e.v]) // (
            e.TermAssociation["t1"] & e.TermAssociation["t2"]
        )
        res = nl.query((e.v, e.p), e.probmap[e.v, e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            (
                "v1",
                (0.6 * 0.1 * 0.2 + 0.4 * 0.3 * 0.4 + 0.2 * 0.5 * 0.6)
                / (0.1 * 0.2 + 0.3 * 0.4 + 0.5 * 0.6),
            ),
            (
                "v2",
                (0.5 * 0.1 * 0.2 + 0.3 * 0.4 * 0.3 + 0.1 * 0.5 * 0.6)
                / (0.1 * 0.2 + 0.3 * 0.4 + 0.5 * 0.6),
            ),
        ]
    )
    assert_almost_equal(res, expected)


def test_cbma_prob_query_with_negation():
    nl = NeurolangPDL()
    nl.add_uniform_probabilistic_choice_over_set(
        [
            ("s1",),
            ("s2",),
            ("s3",),
        ],
        name="SelectedStudy",
    )
    nl.add_tuple_set([("nA",), ("nB",)], name="Network")
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.1, "t1", "s1"),
            (0.2, "t2", "s1"),
            (0.3, "t1", "s2"),
            (0.4, "t2", "s2"),
            (0.5, "t1", "s3"),
            (0.6, "t2", "s3"),
        ],
        name="TermInStudy",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.6, "nA", "s1"),
            (0.5, "nB", "s1"),
            (0.4, "nA", "s2"),
            (0.3, "nB", "s2"),
            (0.2, "nA", "s3"),
            (0.1, "nB", "s3"),
        ],
        name="NetworkReported",
    )
    with nl.environment as e:
        e.ProbTermAssociation[e.t, e.n, e.PROB[e.t, e.n]] = (
            e.TermInStudy[e.t, e.s] & e.SelectedStudy[e.s]
        ) // (
            ~e.NetworkReported[e.n, e.s]
            & e.Network[e.n]
            & e.SelectedStudy[e.s]
        )
        res = nl.query((e.t, e.n, e.p), e.ProbTermAssociation[e.t, e.n, e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            ("t1", "nA", 0.344444),
            ("t1", "nB", 0.338095),
            ("t2", "nA", 0.444444),
            ("t2", "nB", 0.438095),
        ]
    )
    assert_almost_equal(res, expected)


def test_query_based_spatial_prior():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            (5, 5, 5, "s1"),
            (7, 5, 5, "s1"),
            (5, 5, 5, "s2"),
            (10, 10, 10, "s2"),
            (10, 10, 11, "s2"),
        ],
        name="FocusReported",
    )
    nl.add_tuple_set(
        list(itertools.product(range(15), range(15), range(15))), name="Voxel"
    )
    nl.add_uniform_probabilistic_choice_over_set(
        [("s1",), ("s2",)], name="SelectedStudy"
    )
    exp = nl.add_symbol(np.exp, name="exp", type_=Callable[[float], float])
    with nl.environment as e:
        (e.VoxelReported @ e.max(exp(-(e.d ** 2) / 5.0)))[
            e.i1, e.j1, e.k1, e.s
        ] = (
            e.FocusReported(e.i2, e.j2, e.k2, e.s)
            & e.Voxel(e.i1, e.j1, e.k1)
            & (e.d == e.EUCLIDEAN(e.i1, e.j1, e.k1, e.i2, e.j2, e.k2))
            & (e.d < 1)
        )
        e.Activation[e.i, e.j, e.k, e.PROB[e.i, e.j, e.k]] = e.VoxelReported(
            e.i, e.j, e.k, e.s
        ) & e.SelectedStudy(e.s)
        result = nl.query(
            (e.i, e.j, e.k, e.p), e.Activation(e.i, e.j, e.k, e.p)
        )
    expected = RelationalAlgebraFrozenSet(
        [
            (4, 5, 5, np.exp(-1 / 5)),
            (5, 4, 5, np.exp(-1 / 5)),
            (5, 5, 4, np.exp(-1 / 5)),
            (5, 5, 5, 1.0),
            (5, 5, 6, np.exp(-1 / 5)),
            (5, 6, 5, np.exp(-1 / 5)),
            (6, 5, 5, np.exp(-1 / 5)),
            (5, 6, 5, np.exp(-1 / 5)),
            (6, 5, 5, np.exp(-1 / 5)),
            (7, 4, 5, np.exp(-1 / 5) / 2),
            (7, 5, 4, np.exp(-1 / 5) / 2),
            (7, 5, 5, 1 / 2),
            (7, 5, 6, np.exp(-1 / 5) / 2),
            (7, 6, 5, np.exp(-1 / 5) / 2),
            (8, 5, 5, np.exp(-1 / 5) / 2),
            (9, 10, 10, np.exp(-1 / 5) / 2),
            (9, 10, 11, np.exp(-1 / 5) / 2),
            (10, 9, 10, np.exp(-1 / 5) / 2),
            (10, 9, 11, np.exp(-1 / 5) / 2),
            (10, 10, 9, np.exp(-1 / 5) / 2),
            (10, 10, 10, 0.5),
            (10, 10, 11, 0.5),
            (10, 10, 12, np.exp(-1 / 5) / 2),
            (10, 11, 10, np.exp(-1 / 5) / 2),
            (10, 11, 11, np.exp(-1 / 5) / 2),
            (11, 10, 10, np.exp(-1 / 5) / 2),
            (11, 10, 11, np.exp(-1 / 5) / 2),
        ]
    )
    assert_almost_equal(result, expected)


def test_simple_sigmoid():
    nl = NeurolangPDL()
    unnormalised = nl.add_tuple_set(
        [
            ("a", 1),
            ("b", 2),
            ("c", 7),
        ],
        name="unnormalised",
    )
    exp = nl.add_symbol(np.exp, name="exp", type_=Callable[[float], float])
    with nl.environment as e:
        (e.prob @ (1 / (1 + exp(-e.x))))[e.l] = unnormalised[e.l, e.x]
        e.Query[e.l, e.PROB[e.l]] = e.prob[e.l]
        res = nl.query((e.l, e.p), e.Query[e.l, e.p])
    expected = {(l, (1 / (1 + np.exp(-x)))) for l, x in unnormalised.value}
    assert_almost_equal(res, expected)


def test_postprob_conjunct_with_wlq_result():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [("s1",), ("s2",)],
        name="S",
    )
    nl.add_uniform_probabilistic_choice_over_set(
        [("s1",), ("s2",)],
        name="SS",
    )
    nl.add_tuple_set(
        [("t1", "s1"), ("t1", "s2"), ("t2", "s2")],
        name="TIS",
    )

    with nl.environment as e:
        e.TheWLQ[e.t, e.PROB(e.t)] = e.SS(e.s) & e.TIS(e.t, e.s)
        e.P[e.t, e.count(e.s)] = e.TIS(e.t, e.s)
        e.Q[e.count(e.s)] = e.S(e.s)
        e.TheQuery[e.t, e.N, e.m, e.p] = (
            e.TheWLQ(e.t, e.p) & e.P(e.t, e.m) & e.Q(e.N)
        )
        sol = nl.query((e.t, e.N, e.m, e.p), e.TheQuery(e.t, e.N, e.m, e.p))
    expected = {
        ("t1", 2, 2, 1.0),
        ("t2", 2, 1, 0.5),
    }
    assert_almost_equal(sol, expected)


def test_no_tuple_unicity_qbased_pfact():
    nl = NeurolangPDL(check_qbased_pfact_tuple_unicity=True)
    nl.add_tuple_set([(0.2, "a"), (0.5, "b"), (0.9, "a")], name="P")
    with nl.scope as e:
        (e.Q @ e.p)[e.x] = e.P(e.p, e.x)
        e.Query[e.x, e.PROB(e.x)] = e.Q(e.x)
        with pytest.raises(RepeatedTuplesInProbabilisticRelationError):
            nl.query((e.x, e.p), e.Query(e.x, e.p))
    nl = NeurolangPDL(check_qbased_pfact_tuple_unicity=False)
    nl.add_tuple_set([(0.2, "a"), (0.5, "b"), (0.9, "a")], name="P")
    with nl.scope as e:
        (e.Q @ e.p)[e.x] = e.P(e.p, e.x)
        e.Query[e.x, e.PROB(e.x)] = e.Q(e.x)
        try:
            nl.query((e.x, e.p), e.Query(e.x, e.p))
        except NeuroLangException:
            pytest.fail(
                "Expected this test not to raise an exception as "
                "the tuple unicity check on query-based probabilistic tables "
                "is disabled"
            )


def test_qbased_pfact_max_prob():
    nl = NeurolangPDL()
    nl.add_tuple_set([(0.2, "a"), (0.5, "b"), (0.9, "a")], name="P")
    with nl.environment as e:
        (e.Q @ e.max(e.p))[e.x] = e.P(e.p, e.x)
        e.Query[e.x, e.PROB(e.x)] = e.Q(e.x)
        sol = nl.query((e.x, e.p), e.Query(e.x, e.p))
    expected = {("a", 0.9), ("b", 0.5)}
    assert_almost_equal(sol, expected)


def test_qbased_pchoice():
    nl = NeurolangPDL()
    nl.add_tuple_set([(0.2, "a"), (0.5, "b"), (0.3, "c")], name="P")
    nl.add_tuple_set([(1, "a"), (1, "c"), (2, "b")], name="R")
    with nl.environment as e:
        (e.Q ^ e.p)[e.x] = e.P(e.p, e.x)
        e.Query[e.x, e.PROB(e.x)] = e.R(e.x, e.y) & e.Q(e.y)
        sol = nl.query((e.x, e.p), e.Query(e.x, e.p))
    expected = {(1, 0.5), (2, 0.5)}
    assert_almost_equal(sol, expected)


def test_qbased_pchoice_equiprobable():
    nl = NeurolangPDL()
    nl.add_tuple_set([("a",), ("b",), ("c",)], name="P")
    nl.add_tuple_set([(1, "a"), (1, "c"), (2, "b")], name="R")
    with nl.environment as e:
        e.S[e.count(e.x)] = e.P(e.x)
        (e.Q ^ (1. / e.p))[e.x] = e.P(e.x) & e.S(e.p)
        e.Query[e.x, e.PROB(e.x)] = e.R(e.x, e.y) & e.Q(e.y)
        sol = nl.query((e.x, e.p), e.Query(e.x, e.p))
    expected = {(1, 2 / 3.), (2, 1 / 3.)}
    assert_almost_equal(sol, expected)


def test_noisy_or_probabilistic_query():
    nl = NeurolangPDL()
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "a", "b"),
            (0.7, "a", "c"),
            (0.9, "b", "c"),
        ],
        name="R",
    )
    with nl.scope as e:
        e.Query[e.x, e.PROB(e.x)] = e.R(e.x, e.y)
        sol = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 1 - (1 - 0.2) * (1 - 0.7)),
        ("b", 1 - (1 - 0.9)),
    }
    assert_almost_equal(sol, expected)


def test_disjunctive_probfact_rule():
    nl = NeurolangPDL()
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.7, "a"),
            (0.1, "b"),
            (0.8, "c"),
        ],
        name="P",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.4, "a"),
            (0.8, "b"),
        ],
        name="Q",
    )
    with nl.scope as e:
        e.Z[e.x] = e.P(e.x)
        e.Z[e.x] = e.Q(e.x)
        e.Query[e.x, e.PROB(e.x)] = e.Z(e.x)
        sol = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 1 - (1 - 0.7) * (1 - 0.4)),
        ("b", 1 - (1 - 0.1) * (1 - 0.8)),
        ("c", 0.8),
    }
    assert_almost_equal(sol, expected)


def test_probchoice_disjunction():
    nl = NeurolangPDL()
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.7, "a"),
            (0.3, "b"),
        ],
        name="P",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "a"),
            (0.9, "c"),
        ],
        name="Q",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "d", "c"),
            (0.1, "b", "a"),
            (0.4, "b", "b"),
        ],
        name="R",
    )
    with nl.scope as e:
        e.Z[e.x] = e.P(e.x) & e.Q(e.x)
        e.Z[e.x] = e.P(e.x) & e.R(e.x, e.y)
        e.Query[e.x, e.PROB(e.x)] = e.Z(e.x)
        sol = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 0.7 * 0.8),
        ("b", 0.3 * (1 - (1 - 0.1) * (1 - 0.4))),
    }
    assert_almost_equal(sol, expected)


def test_probchoice_disjunction_probfact_probchoice():
    nl = NeurolangPDL()
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.7, "a"),
            (0.3, "b"),
        ],
        name="P",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "a"),
            (0.9, "c"),
        ],
        name="Q",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "d", "c"),
            (0.1, "b", "a"),
            (0.4, "b", "b"),
        ],
        name="R",
    )
    with nl.scope as e:
        e.Z[e.x] = e.P(e.x) & e.Q(e.x)
        e.Z[e.x] = e.R(e.x, e.y)
        e.Query[e.x, e.PROB(e.x)] = e.Z(e.x)
        sol = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 0.7 * 0.8),
        ("b", 1 - (1 - 0.1) * (1 - 0.4)),
        ("d", 0.2),
    }
    assert_almost_equal(sol, expected)


def test_simple_disjunction_probfacts():
    nl = NeurolangPDL()
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.7, "a"),
            (0.3, "b"),
        ],
        name="P",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "a"),
            (0.9, "c"),
        ],
        name="Q",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "d"),
            (0.1, "b"),
        ],
        name="R",
    )
    with nl.scope as e:
        e.Z[e.x] = e.P(e.x) & e.Q(e.x)
        e.Z[e.x] = e.R(e.x)
        e.Query[e.x, e.PROB(e.x)] = e.Z(e.x)
        sol = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 0.7 * 0.8),
        ("b", 0.1),
        ("d", 0.2),
    }
    assert_almost_equal(sol, expected)


def test_pchoice_with_both_eqvar_and_free_var():
    nl = NeurolangPDL()
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.7, "a", "b"),
            (0.2, "a", "c"),
            (0.1, "b", "b"),
        ],
        name="Pchoice",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.9, "a"),
            (0.2, "b"),
            (1.0, "c"),
        ],
        name="Pfact",
    )
    with nl.scope as e:
        e.Query[e.x, e.PROB(e.x)] = e.Pchoice(e.x, e.y) & e.Pfact(e.y)
        result = nl.query((e.x, e.prob), e.Query(e.x, e.prob))
    expected = {
        ("a", 0.7 * 0.2 + 0.2 * 1.0),
        ("b", 0.1 * 0.2),
    }
    assert_almost_equal(result, expected)


def test_current_program_with_probfact():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [
            (10751455, "emotion", 0.052538),
            (10808134, "emotion", 0.244368),
            (10913505, "emotion", 0.059463),
        ],
        name="TermInStudyTFIDF",
    )
    exp = nl.add_symbol(np.exp, name="exp", type_=Callable[[float], float])
    with nl.environment as e:
        (e.TermInStudy @ (1 / (1 + exp(-300 * (e.tfidf - 0.001)))))[
            e.t, e.s
        ] = e.TermInStudyTFIDF(e.s, e.t, e.tfidf)
    prog = nl.current_program
    assert len(prog) == 2



import warnings

import rdflib
from rdflib import BNode, Literal
from rdflib.extras.infixowl import EnumeratedClass
from rdflib.namespace import OWL, RDF, RDFS, SKOS

from ..config import config
from ..exceptions import NeuroLangException, NeuroLangNotImplementedError
from ..expressions import (
    Constant,
    FunctionApplication,
    Symbol,
    sure_is_not_pattern
)
from ..logic import Conjunction, Implication
from .constraints_representation import RightImplication


class OntologyParser:
    """
    This class is in charge of generating the rules that can be derived
    from an ontology, both at entity and constraint levels.
    """

    def __init__(self, paths, connector_symbol=None, load_format="xml"):
        self.structural_knowledge_namespace = (
            config.get_structural_knowledge_namespace()
        )
        if isinstance(paths, list):
            self._load_ontology(paths, load_format)
        else:
            self._load_ontology([paths], [load_format])

        if connector_symbol is None:
            self.connector_symbol = Symbol.fresh()
        else:
            self.connector_symbol = connector_symbol

        self.parsed_constraints = {}
        self.estructural_knowledge = {}
        self.existential_rules = {}
        self.entity_rules = {}

    def _load_ontology(self, paths, load_format):
        rdfGraph = rdflib.Graph()
        for counter, path in enumerate(paths):
            rdfGraph.parse(path, format=load_format[counter])

        self.rdfGraph = rdfGraph

    def parse_ontology(self):
        """
        Main function to be called for ontology processing.

        Returns
        -------
        Two dictionaries. One with contraints, for example, rules derived
        from restrictions of type someValuesFrom. The other one, with the
        implications derived from the rest of the ontological information,
        for example, the label property. Both dictionaries index lists of
        rules using the rule consequent functor.
        """
        self._parse_classes()
        self._parse_related_individuals()

        return (
            self.parsed_constraints,
            self.estructural_knowledge,
            self.entity_rules,
        )

    def _parse_classes(self):
        """This method obtains all the classes present in the ontology and
        iterates over them, parsing them according to the properties
        that compose them.
        """
        all_classes = self._get_all_classes()
        for class_name in list(all_classes):

            for entity, prop, value in self.rdfGraph.triples(
                (class_name, None, None)
            ):

                if prop == RDF.type and value == OWL.Class:
                    # The triple used to index the class must be skipped.
                    continue

                elif prop == RDFS.subClassOf:
                    self._parseSubClass(entity, value)

                elif prop == OWL.equivalentClass:
                    self._parseEquivalentClass(entity, prop, value)

                elif prop == EnumeratedClass:
                    self._parseEnumeratedClass(entity, prop, value)

                else:
                    if not isinstance(value, BNode):
                        self._parse_property(entity, prop, value)
                    else:
                        raise NeuroLangNotImplementedError

    def _get_all_classes(self):
        """
        Function in charge of obtaining all the classes of the ontology
        to iterate through at the time of parsing.

        Returns
        -------
        A set of URIs representing the classes that compose the ontology.
        """
        classes = set()
        for s, _, _ in self.rdfGraph.triples((None, RDF.type, OWL.Class)):
            if not isinstance(s, BNode):
                classes.add(s)
        for s, _, _ in self.rdfGraph.triples((None, RDF.type, RDFS.Class)):
            if not isinstance(s, BNode):
                classes.add(s)

        return classes

    def _get_all_named_individual(self):
        individuals = set()
        for s, _, _ in self.rdfGraph.triples(
            (None, RDF.type, OWL.NamedIndividual)
        ):
            individuals.add(s)

        return individuals

    def _parse_related_individuals(self):
        _all_individuals = self._get_all_named_individual()

        for individual_name in list(_all_individuals):
            for entity, prop, value in self.rdfGraph.triples(
                (individual_name, None, None)
            ):

                if prop == RDF.type and value == OWL.NamedIndividual:
                    # The triple used to index the individual must be skipped.
                    continue

                if prop == SKOS.related:
                    self._parse_individual_related(entity, prop, value)

    def _parseSubClass(self, entity, val):
        """Function in charge of identifying the type of constraint to be
        parsed. At the moment it allows to parse subclasses, basic
        constraints and a single level of intersections.
        Nested intersections are not yet implemented.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        value : URIRef or BNode
            value associated with the entity.
        """
        res = []
        if isinstance(val, BNode):
            bnode_triples = list(self.rdfGraph.triples((val, None, None)))
            restriction_dic = {b: c for _, b, c in bnode_triples}
            if OWL.Restriction in restriction_dic.values():
                res = self._parse_restriction(entity, restriction_dic)
                self._categorize_constraints(res)
            elif OWL.intersectionOf in restriction_dic.keys():
                c = restriction_dic[OWL.intersectionOf]
                int_triples = list(self.rdfGraph.triples((c, None, None)))
                inter_entity = [
                    a[2] for a in int_triples if not isinstance(a[2], BNode)
                ]
                inter_BNode = [
                    a[2] for a in int_triples if isinstance(a[2], BNode)
                ]
                if len(inter_entity) == 1 and len(inter_BNode) == 1:
                    self._parse_BNode_intersection(
                        entity, inter_BNode[0], inter_entity[0]
                    )
                else:
                    warnings.warn(
                        "Complex intersectionOf are not implemented yet"
                    )

            elif OWL.unionOf in restriction_dic.keys():
                warnings.warn("Not implemented yet: unionOf")
            elif OWL.complementOf in restriction_dic.keys():
                warnings.warn("Not implemented yet: complementOf")
            else:
                raise NotImplementedError(
                    f"Something went wrong: {restriction_dic}"
                )
        else:
            ant = Symbol(self._parse_name(entity))
            cons = Symbol(self._parse_name(val))
            x = Symbol.fresh()
            imp = RightImplication(
                FunctionApplication[bool](
                    ant,
                    (x,),
                    verify_type=False
                ),
                FunctionApplication[bool](
                    cons,
                    (x,),
                    verify_type=False
                )
            )
            self._categorize_constraints([imp])

            nl_subclassof = Symbol(
                self.structural_knowledge_namespace + "subClassOf"
            )
            est = FunctionApplication[bool](
                nl_subclassof,
                (
                    Constant[str](ant.name, verify_type=False),
                    Constant[str](cons.name, verify_type=False)
                ),
                verify_type=False
            )
            self._categorize_structural_knowledge([est])

    def _parse_BNode_intersection(self, entity, node, inter_entity):
        """When the rules that compose a constraint are defined within an
        intersection, it needs to be manipulated in a special way.
        This is the method in charge of that behavior.

        Parameters
        ----------
        entity : URIRef
            the main entity containing the intersection.

        node : URIRef
            URI defining the intersection.

        inter_entity : URIRef
            the main entity defined within the intersection.

        """
        triple_restriction = list(self.rdfGraph.triples((node, None, None)))
        nil = [a[2] for a in triple_restriction if not isinstance(a[2], BNode)]
        bnode = [a[2] for a in triple_restriction if isinstance(a[2], BNode)]
        support_prop = Symbol.fresh()
        if nil[0] == RDF.nil:
            bnode_triples = list(self.rdfGraph.triples((bnode[0], None, None)))
            restriction_dic = {b: c for _, b, c in bnode_triples}
            if OWL.Restriction in restriction_dic.values():
                res = self._parse_restriction(
                    entity, restriction_dic, support_prop
                )
                self._categorize_constraints(res)
            con = Symbol(self._parse_name(inter_entity))
            x = Symbol.fresh()
            y = Symbol.fresh()
            imp = RightImplication(
                FunctionApplication[bool](
                    support_prop,
                    (x, y),
                    verify_type=False
                ),
                FunctionApplication[bool](con, (x,), verify_type=False)
            )
            self._categorize_constraints([imp])
        else:
            warnings.warn("Complex intersectionOf are not implemented yet")

    def _parse_name(self, obj):
        """Function that transforms the names of the entities of the ontology
        while preserving the namespace to avoid conflicts.

        Example: the URI `http://www.w3.org/2004/02/skos/core#altLabel`
        becomes `core:altLabel`.


        Parameters
        ----------
        obj : URIRef
            entity to be renamed.

        Returns
        -------
        String with the new name associated to the entity.
        """
        if isinstance(obj, Literal):
            return str(obj)

        obj_split = obj.split("#")
        if len(obj_split) == 2:
            name = obj_split[1]
            namespace = obj_split[0].split("/")[-1]
            if namespace + ":" == self.structural_knowledge_namespace:
                raise Warning(
                    f"The ontology namespace {namespace} matches the one\n"
                    f"used by neurolang to define structural knowledge. It is\n"
                    f"recommended to change the latter using the"
                    f"set_structural_knowledge_namespace method in\n"
                    f"the configuration module to avoid unwanted behavior.\n"
                )
            if name[0] != "" and namespace != "":
                res = namespace + ":" + name
            else:
                res = name
        else:
            obj_split = obj.split("/")
            res = obj_split[-1]

        return res

    def _parse_restriction(self, entity, restriction_dic, support_prop=None):
        """Method for the identification of the type of restriction.
        Each of the possible available restrictions has its own
        method in charge of information parsing.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        restriction_dic : dict
            dictionary containing the information of the constraint
            to be processed.

        support_prop : Symbol, default None
            Optional symbol. It is used to predefine the symbol used in
            rules with existentials, when necessary
            (mainly in nested definitions).

        Returns
        -------
        A list of rules that compose the parsed constraint.

        """
        cons = []
        prop = restriction_dic[OWL.onProperty]

        if OWL.someValuesFrom in restriction_dic.keys():
            node = restriction_dic[OWL.someValuesFrom]
            if isinstance(node, BNode):
                node = self._solve_BNode(node)
            else:
                node = [node]
            cons = self._parse_someValuesFrom(
                entity, prop, node, support_prop=support_prop
            )

        elif OWL.allValuesFrom in restriction_dic.keys():
            node = restriction_dic[OWL.allValuesFrom]
            if isinstance(node, BNode):
                node = self._solve_BNode(node)
            else:
                node = [node]
            cons = self._parse_allValuesFrom(entity, prop, node)

        elif OWL.hasValue in restriction_dic.keys():
            node = restriction_dic[OWL.hasValue]
            if isinstance(node, BNode):
                node = self._solve_BNode(node)
            else:
                node = [node]
            cons = self._parse_hasValue(entity, prop, node)

        elif OWL.minCardinality in restriction_dic.keys():
            warnings.warn(
                "minCardinality constraints cannot\n"
                "be implemented in Datalog syntax"
            )
        elif OWL.maxCardinality in restriction_dic.keys():
            warnings.warn(
                "maxCardinality constraints cannot\n"
                "be implemented in Datalog syntax"
            )
        else:
            raise NeuroLangException(
                "This restriction does not correspond to an OWL DL constraint:",
                restriction_dic.keys(),
            )

        return cons

    def _parse_someValuesFrom(self, entity, prop, nodes, support_prop=None):
        """After the restriction is identified as of type `someValuesFrom`,
        this function is in charge of parsing the information and returning it
        in the form of rules that our Datalog program can interpret.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        prop : URIRef
            propertie associated with the entity.

        nodes : list
            list of the values associated with the entity and the property.

        support_prop : Symbol, default None
            Optional symbol. It's used to predefine the symbol used in
            rules with existentials, when necessary
            (mainly in nested definitions).
        """
        constraints = []
        with sure_is_not_pattern():
            if support_prop is None:
                support_prop = Symbol.fresh()
            x = Symbol.fresh()
            y = Symbol.fresh()
            onProp = Symbol(self._parse_name(prop))
            entity = Symbol(self._parse_name(entity))
            support_prop_x_y = FunctionApplication[bool](
                support_prop,
                (x, y),
                verify_type=False
            )
            entity_x = FunctionApplication[bool](
                entity,
                (x,),
                verify_type=False
            )
            for value in nodes:
                value = self._parse_name(value)
                value_y = FunctionApplication[bool](
                    Symbol(value), (y,), verify_type=False
                )
                exists_imp = RightImplication(support_prop_x_y, value_y)
                constraints.append(exists_imp)
            on_prop_x_y = FunctionApplication[bool](
                onProp,
                (x, y),
                verify_type=False
            )
            prop_imp = RightImplication(support_prop_x_y, on_prop_x_y)
            ext_rule = RightImplication(entity_x, support_prop_x_y)
            self._add_existential_rule(entity_x, ext_rule)

            constraints.append(ext_rule)
            constraints.append(prop_imp)

        return constraints

    def _parse_allValuesFrom(self, entity, prop, nodes):
        """After the restriction is identified as of type `allValuesFrom`,
        this function is in charge of parsing the information and returning it
        in the form of rules that our Datalog program can interpret.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        prop : URIRef
            propertie associated with the entity.

        nodes : list
            list of the values associated with the entity and the property.
        """
        constraints = []
        x = Symbol.fresh()
        y = Symbol.fresh()
        ant = Symbol(self._parse_name(entity))
        onProp = Symbol(self._parse_name(prop))
        conj = Conjunction((ant(x), onProp(x, y)))
        for value in nodes:
            value = self._parse_name(value)
            constraints.append(RightImplication(conj, Symbol(value)(y)))

        return constraints

    def _parse_hasValue(self, entity, prop, nodes):
        """After the constraint is identified as of type `hasValue`,
        this function is in charge of parsing the information and returning it
        in the form of rules that our Datalog program can interpret.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        prop : URIRef
            propertie associated with the entity.

        nodes : list
            list of the values associated with the entity and the property.
        """
        x = Symbol.fresh()
        ent = Symbol(self._parse_name(entity))
        onProp = Symbol(self._parse_name(prop))
        value = self._parse_name(nodes[0])
        return [
            RightImplication(
                Symbol(ent)(x), Symbol(onProp)(
                    x,
                    Constant[str](value, verify_type=False)
                )
            )
        ]

    def _solve_BNode(self, initial_node):
        """Once a BNode is identified, this function iterates
        over each of the pointers that compose it and returns
        a list with those values.

        Parameters
        ----------
        initial_node : BNode
            pointer to the first item in the list

        Returns
        -------
        list
            A list of all the values that arise
            from traversing the list of BNodes.
         """
        list_node = RDF.nil
        values = []
        for node_triples in self.rdfGraph.triples((initial_node, None, None)):
            if OWL.unionOf == node_triples[1]:
                list_node = node_triples[2]

        while list_node != RDF.nil and list_node is not None:
            list_iter = self.rdfGraph.triples((list_node, None, None))
            list_iter = list(list_iter)
            values.append(self._get_list_first_value(list_iter))
            list_node = self._get_list_rest_value(list_iter)

        return values

    def _get_list_first_value(self, list_iter):
        """Used to iterate between pointers in triples, this function takes
        care of finding the pointed value and returning it.

        Parameters
        ----------
        list_iter : list
            list of triples

        Returns
        -------
            see description.
        """
        for triple in list_iter:
            if RDF.first == triple[1]:
                return triple[2]

    def _get_list_rest_value(self, list_iter):
        """Used to iterate between pointers in triples, this function takes
        care of finding the pointer to the next element and returning it.

        Parameters
        ----------
        list_iter : list
            list of triples

        Returns
        -------
            see description.
        """
        for triple in list_iter:
            if RDF.rest == triple[1]:
                return triple[2]

    def _parse_property(self, entity, prop, value):
        """Any rule that is not a restriction or a derivation on
        classes(as the case of subClassOf or EnumeratedClass) is a
        property that defines a field inside the entity with its
        corresponding information. For example, the properties
        label, altLabel, definition, etc.

        This function is in charge of parsing that information.

        Parameters
        ----------
        entity : URIRef
            entity to be parsed.

        prop : URIRef
            propertie associated with the entity.

        nodes : URIRef
            value associated with the entity and the property.
        """
        with sure_is_not_pattern():
            entity = Symbol(self._parse_name(entity))
            entity_name = Constant[str](
                self._parse_name(value), verify_type=False
            )
            x = Symbol.fresh()
            ant = FunctionApplication[bool](
                entity, (x,), verify_type=False
            )
            label = Symbol(self._parse_name(prop))
            con = FunctionApplication[bool](
                label,
                (x, entity_name),
                verify_type=False
            )

            self._categorize_constraints([RightImplication(ant, con)])

            if prop == RDFS.label:
                x = Symbol.fresh()
                lower_name = self._parse_name(value).lower()
                rule = Implication(
                    FunctionApplication[bool](
                        entity,
                        (x,),
                        verify_type=False
                    ),
                    FunctionApplication[bool](
                        self.connector_symbol,
                        (
                            x,
                            Constant[str](lower_name, verify_type=False)
                        ),
                        verify_type=False
                    )
                )
                self._add_rules([rule])

            prop_name = label.name.split(":")[-1]
            neurolang_prop = Symbol(
                self.structural_knowledge_namespace + prop_name
            )
            est = FunctionApplication[bool](
                neurolang_prop,
                (
                    Constant[str](entity.name, verify_type=False),
                    entity_name
                ),
                verify_type=False
            )
            self._categorize_structural_knowledge([est])

    def _parse_individual_related(self, entity, prop, value):
        entity = Symbol(self._parse_name(entity))
        entity_name = Constant[str](self._parse_name(value), verify_type=False)
        label = Symbol(self._parse_name(prop))

        prop_name = label.name.split(":")[-1]
        neurolang_prop = Symbol(
            self.structural_knowledge_namespace + prop_name
        )
        est = FunctionApplication[bool](
            neurolang_prop,
            (
                Constant[str](entity.name, verify_type=False),
                entity_name
            ),
            verify_type=False
        )
        self._categorize_structural_knowledge([est])

    def _parseEnumeratedClass(self, entity, prop, value):
        warnings.warn("Not implemented yet: EnumeratedClass")

    def _parseEquivalentClass(self, entity, prop, value):
        warnings.warn("Not implemented yet: EquivalentClass")

    def _categorize_constraints(self, formulas):
        for sigma in formulas:
            sigma_functor = sigma.consequent.functor.name
            if sigma_functor in self.parsed_constraints:
                cons_set = self.parsed_constraints[sigma_functor]
                cons_set.add(sigma)
                self.parsed_constraints[sigma_functor] = cons_set
            else:
                self.parsed_constraints[sigma_functor] = set([sigma])

    def _categorize_structural_knowledge(self, formulas):
        for sigma in formulas:
            sigma_functor = sigma.functor
            if sigma_functor in self.estructural_knowledge:
                cons_set = self.estructural_knowledge[sigma_functor]
                cons_set.add(sigma)
                self.estructural_knowledge[sigma_functor] = cons_set
            else:
                self.estructural_knowledge[sigma_functor] = set([sigma])

    def _add_existential_rule(self, entity, rule):
        if entity not in self.existential_rules:
            self.existential_rules[entity] = [rule]
        else:
            rules = self.existential_rules[entity]
            rules.append(rule)
            self.existential_rules[entity] = rules

    def _add_rules(self, expressions):
        for exp in expressions:
            exp_functor = exp.consequent.functor
            if exp_functor in self.entity_rules:
                cons_set = self.entity_rules[exp_functor]
                cons_set.add(exp)
                self.entity_rules[exp_functor] = cons_set
            else:
                self.entity_rules[exp_functor] = set([exp])



import logging
import operator
import typing

import numpy
import pandas
import scipy.spatial

from ....datalog.expression_processing import (
    UnifyVariableEqualitiesMixin,
    extract_logic_atoms,
)
from ....exceptions import UnsupportedProgramError
from ....expression_pattern_matching import add_match
from ....expression_walker import IdentityWalker, PatternWalker
from ....expressions import Constant, Expression, FunctionApplication, Symbol
from ....logic import Conjunction, Implication
from ....relational_algebra import RelationalAlgebraSet
from ....utils import log_performance

EUCLIDEAN = Symbol("EUCLIDEAN")

LOG = logging.getLogger(__name__)


class DetectEuclideanDistanceBoundMatrix(PatternWalker):
    """
    Detect a Euclidean spatial bound in the antecedent of a rule.

    This spatial bound is defined by 4 conjuncts in the antecedent of the rule:
        - an equality formula between some variable `d` and an expression of
          the form `EUCLIDEAN(i1, j1, k1, i2, j2, k2)`, where `EUCLIDEAN` is a
          special reserved symbol for the Euclidean distance function in a
          3-dimensional space.
        - a comparison formula between that same variable `d` and a constant
          value that sets an upper bound for the distance between points,
          thereby limiting the set of points that will be considered in the
          computation.
        - a formula of the form `R1(x)` where x is a tuple containing `i1`,
          `j1` and `k1` providing a range for (i1, j1, k1) coordinates
        - a formula of the form `R2(x)` where x is a tuple containing `i2`,
          `j2` and `k2` providing a range for (i2, j2, k2) coordinates

    """

    def __init__(self):
        self._rule_normaliser = self._RuleNormaliser()

    class _RuleNormaliser(
        UnifyVariableEqualitiesMixin,
        IdentityWalker,
    ):
        pass

    @add_match(Implication(FunctionApplication, Conjunction))
    def implication(self, implication):
        implication = self._rule_normaliser.walk(implication)
        formulas = extract_logic_atoms(implication.antecedent)
        var_to_euclidean_equality_formula = self.get_var_to_euclidean_equality(
            formulas
        )
        distance_upper_bound_formula = self.get_distance_upper_bound(formulas)
        if var_to_euclidean_equality_formula is None:
            return False
        elif distance_upper_bound_formula is None:
            raise UnsupportedProgramError(
                "A maximum distance must be provided when defining a spatial "
                "prior using the Euclidean distance between two sets of points"
            )
        i1, j1, k1, i2, j2, k2 = next(
            arg
            for arg in var_to_euclidean_equality_formula.args
            if isinstance(arg, FunctionApplication)
        ).args
        if (
            self.get_range_pred_for_coord(formulas, (i1, j1, k1)) is None
            or self.get_range_pred_for_coord(formulas, (i2, j2, k2)) is None
        ):
            return False
        return True

    @add_match(Expression)
    def any_other_expression(self, expression):
        return False

    @staticmethod
    def get_range_pred_for_coord(formulas, coord_args):
        matching_formulas = list(
            formula
            for formula in formulas
            if isinstance(formula.functor, Symbol)
            and formula.functor != EUCLIDEAN
            and all(
                sum(arg == coord_arg for arg in formula.args) == 1
                for coord_arg in coord_args
            )
        )
        if len(matching_formulas) == 0:
            return None
        return matching_formulas[0]

    @staticmethod
    def get_var_to_euclidean_equality(formulas):
        matching_formulas = list(
            formula
            for formula in formulas
            if isinstance(formula.functor, Constant)
            and formula.functor.value == operator.eq
            and len(formula.args) == 2
            and any(isinstance(arg, Symbol) for arg in formula.args)
            and any(
                isinstance(arg, FunctionApplication) for arg in formula.args
            )
            and next(
                arg
                for arg in formula.args
                if isinstance(arg, FunctionApplication)
            ).functor
            == EUCLIDEAN
            and len(
                next(
                    arg
                    for arg in formula.args
                    if isinstance(arg, FunctionApplication)
                ).args
            )
            == 6
        )
        if len(matching_formulas) != 1:
            return None
        return matching_formulas[0]

    @staticmethod
    def get_distance_upper_bound(formulas):
        matching_formulas = list(
            formula
            for formula in formulas
            if isinstance(formula.functor, Constant)
            and formula.functor.value
            in (operator.lt, operator.le, operator.gt, operator.ge)
            and len(formula.args) == 2
            and any(isinstance(arg, Symbol) for arg in formula.args)
            and any(isinstance(arg, Constant) for arg in formula.args)
        )
        if len(matching_formulas) != 1:
            return None
        return matching_formulas[0]


class TranslateEuclideanDistanceBoundMatrixMixin(PatternWalker):
    @add_match(
        Implication(FunctionApplication, Conjunction),
        DetectEuclideanDistanceBoundMatrix().walk,
    )
    def euclidean_spatial_bound(self, implication):
        formulas = implication.antecedent.formulas
        var_to_euclidean_equality_formula = (
            DetectEuclideanDistanceBoundMatrix.get_var_to_euclidean_equality(
                formulas
            )
        )
        distance_upper_bound_formula = (
            DetectEuclideanDistanceBoundMatrix.get_distance_upper_bound(
                formulas
            )
        )
        d = next(
            arg
            for arg in var_to_euclidean_equality_formula.args
            if isinstance(arg, Symbol)
        )
        i1, j1, k1, i2, j2, k2 = next(
            arg
            for arg in var_to_euclidean_equality_formula.args
            if isinstance(arg, FunctionApplication)
        ).args
        upper_bound = next(
            arg
            for arg in distance_upper_bound_formula.args
            if isinstance(arg, Constant)
        )
        first_range_pred = (
            DetectEuclideanDistanceBoundMatrix.get_range_pred_for_coord(
                formulas, (i1, j1, k1)
            )
        )
        second_range_pred = (
            DetectEuclideanDistanceBoundMatrix.get_range_pred_for_coord(
                formulas, (i2, j2, k2)
            )
        )
        first_coord_set = self.safe_range_pred_to_coord_set(
            first_range_pred, (i1, j1, k1)
        )
        second_coord_set = self.safe_range_pred_to_coord_set(
            second_range_pred, (i2, j2, k2)
        )
        max_dist = self.upper_bound_to_max_dist(upper_bound)
        spatial_bound_solution = self.solve_spatial_bound(
            first_coord_set.as_numpy_array(),
            second_coord_set.as_numpy_array(),
            max_dist,
        )
        new_pred_symb = Symbol.fresh()
        spatial_bound_solution_pred = new_pred_symb(i1, j1, k1, i2, j2, k2, d)
        self.add_extensional_predicate_from_tuples(
            new_pred_symb, spatial_bound_solution
        )
        removed_formulas = {
            var_to_euclidean_equality_formula,
            distance_upper_bound_formula,
        }
        new_formulas = (spatial_bound_solution_pred,) + tuple(
            formula for formula in formulas if formula not in removed_formulas
        )
        new_implication = Implication(
            implication.consequent, Conjunction(new_formulas)
        )
        return self.walk(new_implication)

    @staticmethod
    def upper_bound_to_max_dist(upper_bound):
        val = upper_bound.value
        if isinstance(val, (int, float)):
            return float(val)
        raise ValueError("Unsupported distance expression")

    @staticmethod
    def solve_spatial_bound(
        first_coord_array: numpy.array,
        second_coord_array: numpy.array,
        max_dist: float,
    ) -> numpy.array:
        with log_performance(LOG, "spatial bound resolution"):
            first_ckd_tree = scipy.spatial.cKDTree(first_coord_array)
            second_ckd_tree = scipy.spatial.cKDTree(second_coord_array)
            dist_mat = first_ckd_tree.sparse_distance_matrix(
                second_ckd_tree,
                max_dist,
                output_type="ndarray",
                p=2,
            )
            dist_mat = pandas.DataFrame(dist_mat)
            solution_df = pandas.DataFrame(
                numpy.c_[
                    first_coord_array[dist_mat.iloc[:, 0].values],
                    second_coord_array[dist_mat.iloc[:, 1].values],
                ],
                dtype=int,
            )
            solution_df[len(solution_df.columns)] = numpy.atleast_2d(
                dist_mat.iloc[:, 2].values
            ).T
        return solution_df

    def safe_range_pred_to_coord_set(
        self,
        range_pred: FunctionApplication,
        coord_args: typing.Tuple[Symbol],
    ) -> RelationalAlgebraSet:
        pred_symb = range_pred.functor
        if pred_symb not in self.symbol_table:
            raise NotImplementedError(
                "Cannot solve spatial prior with non-extensional dependency "
                "(yet)"
            )
        ra_set = self.symbol_table[pred_symb].value
        args_as_str = list(
            arg.name for arg in range_pred.args if isinstance(arg, Symbol)
        )
        proj_cols = (
            args_as_str.index(coord_arg.name) for coord_arg in coord_args
        )
        return ra_set.projection(*proj_cols)



from .orderedset import OrderedSet
from .relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet,
    RelationalAlgebraFrozenSet,
    RelationalAlgebraSet,
    RelationalAlgebraStringExpression,
)
from .various import log_performance, powerset

__all__ = [
    "OrderedSet",
    "RelationalAlgebraSet",
    "RelationalAlgebraFrozenSet",
    "NamedRelationalAlgebraFrozenSet",
    "RelationalAlgebraStringExpression",
    "log_performance",
    "powerset",
]



from pathlib import Path
from typing import AbstractSet, Dict, Tuple

import nibabel
import pandas as pd
from nibabel.dataobj_images import DataobjImage
from nilearn.datasets.utils import _fetch_files
from pandas.errors import ParserError

from .exceptions import InvalidCommandExpression, UnsupportedProgramError
from .expression_pattern_matching import add_match
from .expression_walker import PatternWalker
from .expressions import Command, Constant, Symbol
from .regions import EmptyRegion, ExplicitVBR


class CommandsMixin(PatternWalker):
    @add_match(Command(Symbol("load_csv"), ..., ...))
    def load_csv(self, command):
        """
        Process the `.load_csv` command. The load_csv command fetches tabular
        data from a given url (using pandas' `read_csv` method) and loads it
        into a relation with the given Symbol name.

        Usage
        -----
        The `.load_csv` command requires two positional arguments:
            - symbol: str, the name of the symbol to load the data into
            - url: str, the url to fetch the data from
        Other keyword arguments are passed to pandas' `read_csv` method with
        limited support (only string values).

        `.load_csv(Study, "https://github.com/neuroquery/neuroquery_data/raw/
        master/data/data-neuroquery_version-1_metadata.tsv.gz", sep="\t")`
        will load NeuroQuery's metadata table into the Study relation

        Raises
        ------
        InvalidCommandExpression
            raised if command args are invalid or if an error occurs while
            loading the data
        """
        try:
            symbol, url = command.args
            if not isinstance(symbol, Symbol):
                symbol = Symbol(symbol)
            url = _unwrap_expr(url)
            kwargs = {
                _unwrap_expr(k): _unwrap_expr(v) for k, v in command.kwargs
            }
        except ValueError:
            raise InvalidCommandExpression(
                "Could not extract Symbol name and url from arguments."
            )

        try:
            data = pd.read_csv(url, **kwargs)
        except ParserError as e:
            raise InvalidCommandExpression(
                f"An error occured while parsing data from {url}: "
            ) from e
        data = data.rename(columns={n: i for i, n in enumerate(data.columns)})
        self.add_extensional_predicate_from_tuples(symbol, data)

    @add_match(Command(Symbol("load_atlas"), ..., ...))
    def load_atlas(self, command):
        """
        Process the `.load_atlas` command. The load_atlas command fetches an
        atlas image with its corresponding labels and loads it into the
        neurolang instance.

        Usage
        -----
        The `.load_csv` command requires three positional arguments:
            - symbol: str, the name of the symbol to load the data into
            - atlas_url: str, the url for the atlas file
            - labels_url: str, the url for the atlas labels
            - base_url: str, optinal. if given, atlas and labels url are
                        considered relative to this base_url
        Other keyword arguments are passed to pandas' `read_csv` method.

        `.load_atlas(destrieux, "destrieux2009_rois_lateralized.nii.gz",
        "destrieux2009_rois_labels_lateralized.csv",
        "https://www.nitrc.org/frs/download.php/11942/destrieux2009.tgz")`

        will load Destrieux's atlas into the destrieux relation.

        Raises
        ------
        InvalidCommandExpression
            raised if command args are invalid or if an error occurs while
            loading the data
        """
        try:
            if len(command.args) == 4:
                symbol, atlas_url, labels_url, base_url = command.args
                base_url = _unwrap_expr(base_url)
            else:
                symbol, atlas_url, labels_url = command.args
                base_url = None
            if not isinstance(symbol, Symbol):
                symbol = Symbol(symbol)
            atlas_url = _unwrap_expr(atlas_url)
            labels_url = _unwrap_expr(labels_url)
            kwargs = {
                _unwrap_expr(k): _unwrap_expr(v) for k, v in command.kwargs
            }
        except ValueError:
            raise InvalidCommandExpression(
                "Could not extract Symbol name and url from arguments."
            )

        atlas_labels, spatial_image = self._fetch_atlas_map_and_labels(
            atlas_url, labels_url, base_url, kwargs
        )

        self._add_atlas_set(symbol, atlas_labels, spatial_image)

    def _fetch_atlas_map_and_labels(
        self, atlas_url, labels_url, base_url, kwargs
    ):
        data_dir = Path.home() / "neurolang_data"
        if base_url is None:
            opts = {}
            p = Path(atlas_url)
            files = [
                (p.name, p.parent, opts),
            ]
            p = Path(labels_url)
            files.append((p.name, p.parent, opts),)
        else:
            opts = {"uncompress": True}
            files = [(atlas_url, base_url, opts), (labels_url, base_url, opts)]

        files_ = _fetch_files(data_dir, files)
        atlas_labels = pd.read_csv(files_[1], **kwargs)
        atlas_labels = {
            label: name for label, name in atlas_labels.to_records(index=False)
        }
        spatial_image = nibabel.load(files_[0])
        return atlas_labels, spatial_image

    def _add_atlas_set(
        self,
        symbol: Symbol,
        atlas_labels: Dict[int, str],
        spatial_image: DataobjImage,
    ) -> None:
        atlas_set = set()
        for idx, label_name in atlas_labels.items():
            region = (ExplicitVBR.from_spatial_image_label(spatial_image, idx),)
            if isinstance(region, EmptyRegion):
                continue
            atlas_set.add((label_name, region))

        type_ = Tuple[str, ExplicitVBR]
        self.add_extensional_predicate_from_tuples(
            Symbol[AbstractSet[type_]](symbol.name), atlas_set, type_=type_
        )

    @add_match(Command)
    def _unknown_command(self, command):
        raise UnsupportedProgramError(
            f"The command statement {command} is not supported."
        )


def _unwrap_expr(expr):
    if isinstance(expr, Constant):
        return expr.value
    return expr.name



import typing

from ...datalog.expression_processing import EQ
from ...expressions import Constant, Symbol
from ...logic import Conjunction
from ..small_dichotomy_theorem_based_solver import (
    is_hierarchical_without_self_joins,
)

Q = Symbol("Q")
R = Symbol("R")
S = Symbol("S")
T = Symbol("T")
w = Symbol("w")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")


def test_hierarchical_without_self_joins():
    q1 = Conjunction((Q(x), R(y), T(x)))
    q2 = Conjunction((Q(x),))
    q3 = Conjunction((Q(x), R(x, y)))
    q4 = Conjunction((Q(x), R(x, y, z), T(x, y)))

    assert is_hierarchical_without_self_joins(q1)
    assert is_hierarchical_without_self_joins(q2)
    assert is_hierarchical_without_self_joins(q3)
    assert is_hierarchical_without_self_joins(q4)


def test_not_hierarchical_without_self_joins():
    q1 = Conjunction((Q(x), T(x, y), R(y)))
    q2 = Conjunction((Q(x), Q(y), T(x, y)))

    assert not is_hierarchical_without_self_joins(q1)
    assert not is_hierarchical_without_self_joins(q2)


def test_hierarchical_despite_multiple_equalities():
    q = Conjunction((Q(x), R(y), EQ(y, z), EQ(x, w)))
    assert is_hierarchical_without_self_joins(q)


def test_non_hierarchical_builtin():
    some_builtin = Constant[typing.Callable[[typing.Any, typing.Any], bool]](
        "some_builtin",
        verify_type=False,
        auto_infer_type=False,
    )
    q = Conjunction((Q(x), R(y), some_builtin(y, z), some_builtin(x, w)))
    assert not is_hierarchical_without_self_joins(q)



# %%
"""
Example where a spatial prior is defined based on the distance between voxels \
and foci in a coordinate-based meta-analysis database
============================================================================

Each voxel's probability of being reported by a study is calculated based on
whether that particular study reports a focus (peak activation) near the voxel.
The probability is defined based on how far from the focus that voxel happens
to be.

"""
import warnings

warnings.filterwarnings("ignore")

from pathlib import Path
from typing import Iterable

import nibabel
import nilearn.datasets
import nilearn.image
import nilearn.plotting
import numpy as np
from neurolang.frontend import ExplicitVBR, ExplicitVBROverlay, NeurolangPDL
from neurolang.frontend.neurosynth_utils import get_ns_mni_peaks_reported

# ##############################################################################
# Data preparation
# ----------------

data_dir = Path.home() / "neurolang_data"

# ##############################################################################
# Load the MNI atlas and resample it to 4mm voxels

mni_t1 = nibabel.load(
    nilearn.datasets.fetch_icbm152_2009(data_dir=str(data_dir / "icbm"))["t1"]
)
mni_t1_2mm = nilearn.image.resample_img(mni_t1, np.eye(3) * 2)

# ##############################################################################
# Probabilistic Logic Programming in NeuroLang
# --------------------------------------------

nl = NeurolangPDL()


# ##############################################################################
# Adding new aggregation function to build a region overlay
# ----------------------------------


@nl.add_symbol
def agg_create_region_overlay(
    i: Iterable, j: Iterable, k: Iterable, p: Iterable
) -> ExplicitVBR:
    voxels = np.c_[i, j, k]
    return ExplicitVBROverlay(
        voxels, mni_t1_2mm.affine, p, image_dim=mni_t1_2mm.shape
    )


# ##############################################################################
# Loading the database
# ----------------------------------

peak_data = get_ns_mni_peaks_reported(data_dir)
ijk_positions = np.round(
    nibabel.affines.apply_affine(
        np.linalg.inv(mni_t1_2mm.affine),
        peak_data[["x", "y", "z"]].values.astype(float),
    )
).astype(int)
peak_data["i"] = ijk_positions[:, 0]
peak_data["j"] = ijk_positions[:, 1]
peak_data["k"] = ijk_positions[:, 2]
peak_data = peak_data[["i", "j", "k", "id"]]

nl.add_tuple_set(peak_data, name="FocusReported")
study_ids = nl.load_neurosynth_study_ids(data_dir, "Study")
nl.add_uniform_probabilistic_choice_over_set(
    study_ids.value, name="SelectedStudy"
)
nl.load_neurosynth_term_study_associations(
    data_dir, "TermInStudyTFIDF", tfidf_threshold=1e-3
)
Voxel = nl.add_tuple_set(
    np.hstack(
        np.meshgrid(
            *(np.arange(0, dim) for dim in mni_t1_2mm.get_fdata().shape)
        )
    )
    .swapaxes(0, 1)
    .reshape(3, -1)
    .T,
    name="Voxel",
)

# ##############################################################################
# Probabilistic program and querying
# ----------------------------------


with nl.environment as e:
    (e.VoxelReported @ e.max(e.exp(-e.d / 5.0)))[e.i1, e.j1, e.k1, e.s] = (
        e.FocusReported(e.i2, e.j2, e.k2, e.s)
        & e.Voxel(e.i1, e.j1, e.k1)
        & (e.d == e.EUCLIDEAN(e.i1, e.j1, e.k1, e.i2, e.j2, e.k2))
        & (e.d < 1)
    )
    e.TermAssociation[e.t] = (
        e.SelectedStudy[e.s] & e.TermInStudyTFIDF[e.s, e.t, ...]
    )
    e.Activation[e.i, e.j, e.k] = (
        e.SelectedStudy[e.s] & e.VoxelReported[e.i, e.j, e.k, e.s]
    )
    e.probmap[e.i, e.j, e.k, e.PROB[e.i, e.j, e.k]] = (
        e.Activation[e.i, e.j, e.k]
    ) // e.TermAssociation["emotion"]
    e.img[e.agg_create_region_overlay[e.i, e.j, e.k, e.p]] = e.probmap[
        e.i, e.j, e.k, e.p
    ]
    img_query = nl.query((e.x,), e.img[e.x])

# ##############################################################################
# Plotting results
# --------------------------------------------

result_image = img_query.fetch_one()[0].spatial_image()
img = result_image.get_fdata()
plot = nilearn.plotting.plot_stat_map(
    result_image, threshold=np.percentile(img[img > 0], 95)
)
nilearn.plotting.show()



from operator import eq
from typing import AbstractSet

import pytest

from ...datalog.translate_to_named_ra import TranslateToNamedRA
from ...expressions import Constant, Symbol
from ...logic import (
    Conjunction,
    Disjunction,
    ExistentialPredicate,
    Implication,
    Negation
)
from ...relational_algebra import (
    ColumnInt,
    ColumnStr,
    Difference,
    NameColumns,
    Projection,
    str2columnstr_constant
)
from ...relational_algebra_provenance import (
    DisjointProjection,
    IndependentProjection
)
from ...utils.relational_algebra_set import NamedRelationalAlgebraFrozenSet
from .. import dalvi_suciu_lift, transforms
from ..probabilistic_ra_utils import (
    DeterministicFactSet,
    ProbabilisticChoiceSet,
    ProbabilisticFactSet
)
from ..transforms import convert_rule_to_ucq

TNRA = TranslateToNamedRA()
EQ = Constant(eq)


def test_has_separator_variable_existential():
    A = Symbol('A')
    B = Symbol('B')
    C = Symbol('C')
    x = Symbol('x')
    y = Symbol('y')

    expression = ExistentialPredicate(
        y,
        Disjunction((B(x, y), Conjunction((A(y), B(x, y))), C(y)))
    )

    sv, _ = dalvi_suciu_lift.find_separator_variables(expression, {})
    assert sv == {y}

    expression = ExistentialPredicate(
        y,
        Disjunction((B(x, y), Conjunction((A(y), B(y, x))), C(y)))
    )

    sv, _ = dalvi_suciu_lift.find_separator_variables(expression, {})
    assert sv == set()

    expression = Disjunction((B(x, y), Conjunction((A(y), B(x, y))), C(y)))

    sv, _ = dalvi_suciu_lift.find_separator_variables(expression, {})
    assert sv == set()

    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')

    expression = ExistentialPredicate(x1, ExistentialPredicate(
        x2,
        ExistentialPredicate(y1, ExistentialPredicate(
            y2,
            Disjunction((
                Conjunction((R(x1), S(x1, y1))),
                Conjunction((T(x2), S(x2, y2)))
            ))
        ))
    ))

    assert dalvi_suciu_lift.has_separator_variables(expression, {})
    assert (
        dalvi_suciu_lift.find_separator_variables(expression, {})[0] &
        {x1, x2}
    )


def test_lifted_bcq_fig_4_4():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    z = Symbol('z')

    cq = Implication(Q(z), Conjunction((
         R(z, x1), S(x1, y1), T(z, x2), S(x2, y2)
    )))
    cq = convert_rule_to_ucq(cq)
    plan = dalvi_suciu_lift.dalvi_suciu_lift(cq, {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert res


def test_lifted_separator_variable():
    R = Symbol('R')
    x1 = Symbol('x1')
    z = Symbol('z')

    cq = ExistentialPredicate(x1, R(x1, z))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(cq, {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert res


def test_lifted_join():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    z = Symbol('z')

    cq = Implication(Q(z), Conjunction((R(x1, z), S(x1, x2), T(x1, z))))
    ucq = convert_rule_to_ucq(cq)
    plan = dalvi_suciu_lift.dalvi_suciu_lift(ucq, {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert res


def test_another_liftable_join():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    z = Symbol('z')

    cq = Implication(Q(z), Conjunction((R(x1, z), S(x1, x2), S(x1, z))))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(convert_rule_to_ucq(cq), {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert res


def test_non_liftable_join():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    z = Symbol('z')

    cq = Implication(Q(z), Conjunction((R(x1, z), S(x2, x1), S(x1, z))))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(convert_rule_to_ucq(cq), {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert not res


def test_lifted_cq_fig_4_5():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    x3 = Symbol('x3')
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    y3 = Symbol('y3')
    z = Symbol('z')

    cq = Implication(Q(z), Disjunction((
        Conjunction((R(z, x1), S(x1, y1))),
        Conjunction((S(x2, y2), T(z, y2))),
        Conjunction((R(z, x3), T(z, y3))),
    )))

    plan = dalvi_suciu_lift.dalvi_suciu_lift(convert_rule_to_ucq(cq), {})
    res = dalvi_suciu_lift.is_pure_lifted_plan(plan)
    assert res


def test_minimize_component_conjunction():
    Q = Symbol('Q')
    R = Symbol('R')
    T = Symbol('T')
    x1 = Symbol('x1')
    y1 = Symbol('y1')
    y2 = Symbol('y2')

    expr = Conjunction((Q(x1),))
    res = transforms.minimize_component_conjunction(expr)
    assert res == expr

    expr = Conjunction((Q(x1, y1), R(x1, y2)))
    res = transforms.minimize_component_conjunction(expr)
    assert res == expr

    expr = Conjunction((
        ExistentialPredicate(y1, Q(x1, y1)),
        ExistentialPredicate(y2, Q(x1, y2))
    ))
    res = transforms.minimize_component_conjunction(expr)
    assert res in (
        Conjunction((q,)) for q in
        (
            ExistentialPredicate(y1, Q(x1, y1)),
            ExistentialPredicate(y2, Q(x1, y2))
        )
    )

    expr = Conjunction((
        ExistentialPredicate(y2, Q(x1, y2)),
        Disjunction((
            ExistentialPredicate(y1, Q(x1, y1)),
            T(x1)
        ))
    ))
    res = transforms.minimize_component_conjunction(expr)
    assert res == Conjunction((ExistentialPredicate(y2, Q(x1, y2)),))

    expr = Conjunction((
        Disjunction((
            ExistentialPredicate(y1, Q(x1, y1)),
            T(x1)
        )),
        ExistentialPredicate(y2, Q(x1, y2)),
    ))
    res = transforms.minimize_component_conjunction(expr)
    assert res == Conjunction((ExistentialPredicate(y2, Q(x1, y2)),))


def test_minimize_component_disjunction():
    Q = Symbol('Q')
    R = Symbol('R')
    T = Symbol('T')
    x1 = Symbol('x1')
    y1 = Symbol('y1')
    y2 = Symbol('y2')

    expr = Disjunction((Q(x1),))
    res = transforms.minimize_component_disjunction(expr)
    assert res == expr

    expr = Disjunction((Q(x1, y1), R(x1, y2)))
    res = transforms.minimize_component_disjunction(expr)
    assert res == expr

    expr = Disjunction((
        ExistentialPredicate(y1, Q(x1, y1)),
        ExistentialPredicate(y2, Q(x1, y2))
    ))
    res = transforms.minimize_component_disjunction(expr)
    assert res in (
        Disjunction((q,)) for q in
        (
            ExistentialPredicate(y1, Q(x1, y1)),
            ExistentialPredicate(y2, Q(x1, y2))
        )
    )

    expr = Disjunction((
        ExistentialPredicate(y2, Q(x1, y2)),
        Disjunction((
            ExistentialPredicate(y1, Q(x1, y1)),
            T(x1)
        ))
    ))
    res = transforms.minimize_component_disjunction(expr)
    assert res == Disjunction((
        ExistentialPredicate(y1, Q(x1, y1)),
        T(x1)
        ))

    expr = Disjunction((
        Disjunction((
            ExistentialPredicate(y1, Q(x1, y1)),
            T(x1)
        )),
        ExistentialPredicate(y2, Q(x1, y2)),
    ))
    res = transforms.minimize_component_disjunction(expr)
    assert res == Disjunction((
        ExistentialPredicate(y1, Q(x1, y1)),
        T(x1)
    ))


def test_intractable_queries():
    """
    Queries that are "intractable", in that they are #P-hard, as defined in
    section 3.2 of [1]_.

    The lifted query processing algorithm should return a NonLiftable object
    containing the unprocessed query.

    [1] Suciu, D. Probabilistic Databases for All. in Proceedings of the
    39th ACM SIGMOD-SIGACT-SIGAI Symposium on Principles of Database Systems
    19–31 (ACM, 2020).
    """
    R = Symbol('R')
    S = Symbol('S')
    S1 = Symbol('S1')
    S2 = Symbol('S2')
    S3 = Symbol('S3')
    T = Symbol('T')
    x = Symbol("x")
    x0 = Symbol('x0')
    x1 = Symbol('x1')
    x2 = Symbol('x2')
    x3 = Symbol('x3')
    y = Symbol("y")
    y0 = Symbol("y0")
    y1 = Symbol('y1')
    y2 = Symbol('y2')
    y3 = Symbol('y3')
    h0 = ExistentialPredicate(
        x,
        Conjunction(
            (
                R(x),
                ExistentialPredicate(y, Conjunction((S(x, y), T(y)))),
            )
        ),
    )
    h1 = Disjunction(
        (
            ExistentialPredicate(
                x0, Conjunction((R(x0), ExistentialPredicate(y0, S(x0, y0))))
            ),
            ExistentialPredicate(
                y1, Conjunction((ExistentialPredicate(x1, S(x1, y1)), T(y1)))
            ),
        )
    )
    h2 = Disjunction(
        (
            ExistentialPredicate(
                x0,
                Conjunction((R(x0), S1(x0, y0))),
            ),
            ExistentialPredicate(
                x1,
                ExistentialPredicate(
                    y1,
                    Conjunction((S1(x1, y1), S2(x1, y1))),
                )
            ),
            ExistentialPredicate(
                y2,
                Conjunction((ExistentialPredicate(x2, S2(x2, y2)), T(y2))),
            ),
        )
    )
    h3 = Disjunction(
        (
            ExistentialPredicate(
                x0,
                Conjunction((R(x0), S1(x0, y0))),
            ),
            ExistentialPredicate(
                x1,
                ExistentialPredicate(
                    y1,
                    Conjunction((S1(x1, y1), S2(x1, y1))),
                )
            ),
            ExistentialPredicate(
                x2,
                ExistentialPredicate(
                    y2,
                    Conjunction((S2(x2, y2), S3(x2, y2))),
                )
            ),
            ExistentialPredicate(
                y3,
                Conjunction((ExistentialPredicate(x3, S3(x3, y3)), T(y3))),
            ),
        )
    )
    for h in (h0, h1, h2, h3):
        plan = dalvi_suciu_lift.dalvi_suciu_lift(h, {})
        assert not dalvi_suciu_lift.is_pure_lifted_plan(plan)


def test_example_4_6_a_really_simple_query():
    R = Symbol("R")
    S = Symbol("S")
    x = Symbol("x")
    y = Symbol("y")
    query = ExistentialPredicate(
        x, Conjunction((R(x), ExistentialPredicate(y, S(x, y))))
    )
    resulting_plan = dalvi_suciu_lift.dalvi_suciu_lift(query, {})
    assert dalvi_suciu_lift.is_pure_lifted_plan(resulting_plan)


def test_example_4_7_a_query_with_self_joins():
    R = Symbol("R")
    S = Symbol("S")
    T = Symbol("T")
    x1 = Symbol("x1")
    x2 = Symbol("x2")
    y1 = Symbol("y1")
    y2 = Symbol("y2")
    Q1 = ExistentialPredicate(
        x1,
        ExistentialPredicate(
            y1,
            Conjunction((R(x1), S(x1, y1))),
        ),
    )
    Q2 = ExistentialPredicate(
        x2,
        ExistentialPredicate(
            y2,
            Conjunction((T(x2), S(x2, y2))),
        ),
    )
    query = Conjunction((Q1, Q2))
    resulting_plan = dalvi_suciu_lift.dalvi_suciu_lift(query, {})
    assert dalvi_suciu_lift.is_pure_lifted_plan(resulting_plan)


@pytest.mark.skip
def test_example_4_8_tractable_query_intractable_subquery():
    """
    We test the query

        R(x1), S(x1, y1) ∨ S(x2, y2), T(y2) ∨ R(x3), T(y3)

    whose first two disjuncts correspond to the hard H1 query, but the third
    disjunct makes the query tractable by using distributivity and logical
    equivalence.

    """
    R = Symbol("R")
    S = Symbol("S")
    T = Symbol("T")
    x1 = Symbol("x1")
    x2 = Symbol("x2")
    x3 = Symbol("x3")
    y1 = Symbol("y1")
    y2 = Symbol("y2")
    y3 = Symbol("y3")
    query = Disjunction(
        (
            ExistentialPredicate(
                x1,
                Conjunction(
                    (
                        R(x1),
                        ExistentialPredicate(y1, S(x1, y1)),
                    )
                ),
            ),
            ExistentialPredicate(
                y2,
                Conjunction(
                    (
                        ExistentialPredicate(x2, S(x2, y2)),
                        T(y2),
                    )
                ),
            ),
            Conjunction(
                (
                    ExistentialPredicate(x3, R(x3)),
                    ExistentialPredicate(y3, T(y3)),
                )
            ),
        )
    )
    resulting_plan = dalvi_suciu_lift.dalvi_suciu_lift(query, {})
    assert dalvi_suciu_lift.is_pure_lifted_plan(resulting_plan)


def test_single_disjoint_project_one_variable():
    P = Symbol("P")
    x = Symbol("x")
    pchoice_relation = NamedRelationalAlgebraFrozenSet(
        iterable=[
            (0.2, "a"),
            (0.7, "b"),
            (0.1, "c"),
        ],
        columns=("_p_", "x"),
    )
    symbol_table = {
        P: ProbabilisticChoiceSet(
            Constant[AbstractSet](pchoice_relation),
            str2columnstr_constant("_p_"),
        ),
    }
    query = ExistentialPredicate(x, P(x))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert isinstance(plan, DisjointProjection)


def test_single_disjoint_project_two_variables():
    P = Symbol("P")
    x = Symbol("x")
    y = Symbol("y")
    pchoice_relation = NamedRelationalAlgebraFrozenSet(
        iterable=[
            (0.2, "a", "b"),
            (0.7, "b", "b"),
            (0.1, "b", "c"),
        ],
        columns=("_p_", "x", "y"),
    )
    symbol_table = {
        P: ProbabilisticChoiceSet(
            Constant[AbstractSet](pchoice_relation),
            str2columnstr_constant("_p_"),
        ),
    }
    query = ExistentialPredicate(y, P(x, y))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert isinstance(plan, DisjointProjection)


def test_simple_existential_query_plan():
    R = Symbol("R")
    x = Symbol("x")
    y = Symbol("y")
    relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                ("a", "b"),
                ("b", "a"),
                ("c", "a"),
            ],
            columns=("x", "y"),
        )
    )
    symbol_table = {R: DeterministicFactSet(relation)}
    query = ExistentialPredicate(x, R(x, y))
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert dalvi_suciu_lift.is_pure_lifted_plan(plan)


def test_extract_probabilistic_root_variables_no_probabilistic_atom():
    res = dalvi_suciu_lift.extract_probabilistic_root_variables(set(), dict())
    assert len(res) == 0

    P = Symbol("P")
    x = Symbol("x")
    y = Symbol("y")
    formulas = [
        P(x, y),
    ]
    relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(
            iterable=[
                ("a", "b"),
                ("b", "a"),
                ("c", "a"),
            ],
            columns=("x", "y"),
        )
    )
    symbol_table = {
        P: DeterministicFactSet(relation)
    }
    res = dalvi_suciu_lift.extract_probabilistic_root_variables(
        formulas, symbol_table
    )
    assert len(res) == 0


def test_lifted_negation():
    R = Symbol('R')
    S = Symbol('S')
    x = Symbol('x')
    y = Symbol('y')

    query = ExistentialPredicate(
        y,
        Conjunction((
            R(x, y),
            Negation(S(y))
        ))
    )

    symbol_table = {
        symbol: ProbabilisticFactSet(Symbol.fresh(), 'p')
        for symbol in (R, S)
    }
    res = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    expected = IndependentProjection(
        Difference(
            NameColumns(
                Projection(
                    R,
                    (Constant(ColumnInt(0)), Constant(ColumnInt(1))),
                ),
                (Constant(ColumnStr('x')), Constant(ColumnStr('y'))),
            ),
            NameColumns(
                Projection(
                    S,
                    (Constant(ColumnInt(0)),)
                ),
                (Constant(ColumnStr('y')),)
            ),
        ),
        (Constant(ColumnStr('x')),)
    )

    assert res == expected


def test_lifted_disjunction_with_negation():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    query = Disjunction((
        ExistentialPredicate(
            y,
            Conjunction((
                R(x, y),
                Negation(S(y))
            ))
        ),
        ExistentialPredicate(
            y,
            T(x, y)
        )
    ))

    symbol_table = {
        symbol: ProbabilisticFactSet(Symbol.fresh(), 'p')
        for symbol in (R, S, T)
    }
    res = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)

    assert dalvi_suciu_lift.is_pure_lifted_plan(res)


def test_lifted_disjunction_cross_product_with_negation():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    V = Symbol('V')
    x = Symbol('x')
    y = Symbol('y')

    query = Disjunction((
        ExistentialPredicate(
            y,
            Conjunction((
                R(x),
                Negation(S(y)),
                V(y)
            ))
        ),
        ExistentialPredicate(
            y,
            T(x, y)
        )
    ))

    symbol_table = {
        symbol: ProbabilisticFactSet(Symbol.fresh(), 'p')
        for symbol in (R, S, T, V)
    }
    res = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)

    assert dalvi_suciu_lift.is_pure_lifted_plan(res)


def test_lifted_conjunction_existential_negation():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    x = Symbol('x')
    y = Symbol('y')

    query = Conjunction((
        Negation(
            ExistentialPredicate(
                x,
                Conjunction((
                    R(x),
                    S(x, y),
                ))
            )
        ),
        T(y)
    ))

    symbol_table = {
        symbol: ProbabilisticFactSet(Symbol.fresh(), 'p')
        for symbol in (R, S, T)
    }
    res = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)

    assert dalvi_suciu_lift.is_pure_lifted_plan(res)


def test_lifted_conjunction_existential_negation_constant():
    R = Symbol('R')
    S = Symbol('S')
    T = Symbol('T')
    Q = Symbol('Q')
    x = Symbol('x')
    y = Symbol('y')

    query = Conjunction((
        Negation(
            ExistentialPredicate(
                x,
                Conjunction((
                    R(x),
                    S(x, y, Constant('2'))
                ))
            )
        ),
        Q(x, y),
        T(y)
    ))

    symbol_table = {
        symbol: ProbabilisticFactSet(Symbol.fresh(), 'p')
        for symbol in (R, S, T)
    }
    symbol_table[Q] = DeterministicFactSet(Symbol.fresh())

    res = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)

    assert dalvi_suciu_lift.is_pure_lifted_plan(res)


def test_mixed_probabilistic_deterministic_safe():
    R = Symbol("R")
    S = Symbol("S")
    T = Symbol("T")
    x = Symbol("x")
    y = Symbol("y")
    R_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("x",))
    )
    S_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "x", "y"))
    )
    T_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "x",))
    )
    symbol_table = {
        R: DeterministicFactSet(R_relation),
        S: ProbabilisticFactSet(S_relation, Constant(ColumnStr("_p_"))),
        T: ProbabilisticFactSet(T_relation, Constant(ColumnStr("_p_"))),
    }
    query = ExistentialPredicate(
        x,
        ExistentialPredicate(
            y,
            Conjunction((R(x), S(x, y), T(y))),
        )
    )
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert dalvi_suciu_lift.is_pure_lifted_plan(plan)


def test_mixed_probabilistic_deterministic_unsafe():
    R = Symbol("R")
    S = Symbol("S")
    T = Symbol("T")
    x = Symbol("x")
    y = Symbol("y")
    R_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("x",))
    )
    S_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "x", "y"))
    )
    T_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "x",))
    )
    symbol_table = {
        R: ProbabilisticFactSet(R_relation, Constant(ColumnStr("_p_"))),
        S: DeterministicFactSet(S_relation),
        T: ProbabilisticFactSet(T_relation, Constant(ColumnStr("_p_"))),
    }
    query = ExistentialPredicate(
        x,
        ExistentialPredicate(
            y,
            Conjunction((R(x), S(x, y), T(y))),
        )
    )
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert not dalvi_suciu_lift.is_pure_lifted_plan(plan)


def test_interaction_deterministic_probfact_probchoice_safe():
    Det = Symbol("Det")
    Pchoice = Symbol("Pchoice")
    Pfact = Symbol("Pfact")
    x = Symbol("x")
    y = Symbol("y")
    Det_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("x",))
    )
    Pchoice_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "y"))
    )
    Pfact_relation = Constant[AbstractSet](
        NamedRelationalAlgebraFrozenSet(iterable=[], columns=("_p_", "x",))
    )
    symbol_table = {
        Det: ProbabilisticFactSet(Det_relation, Constant(ColumnStr("_p_"))),
        Pchoice: DeterministicFactSet(Pchoice_relation),
        Pfact: ProbabilisticFactSet(
            Pfact_relation, Constant(ColumnStr("_p_"))
        ),
    }
    query = ExistentialPredicate(
        x,
        ExistentialPredicate(
            y,
            Conjunction((Det(x, y), Pchoice(y), Pfact(x))),
        )
    )
    plan = dalvi_suciu_lift.dalvi_suciu_lift(query, symbol_table)
    assert dalvi_suciu_lift.is_pure_lifted_plan(plan)



from abc import abstractmethod, abstractproperty
from contextlib import contextmanager
from multiprocessing import BoundedSemaphore
from pathlib import Path
from typing import Callable, Iterable, Union

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import nibabel as nib
import numpy as np
import pandas as pd
from neurolang.frontend import NeurolangDL, NeurolangPDL
from neurolang.frontend.neurosynth_utils import StudyID
from neurolang.regions import (
    ExplicitVBR,
    ExplicitVBROverlay,
    region_union,
)
from nilearn import datasets, image


class NeurolangEngineSet:
    """
    Utility class to hold a set of Neurolang engines (either NeurolangDL
    or NeurolangPDL). It keeps track of the total number of engines added
    to the set, as well as a semaphore to manage acquiring and releasing the
    engines in a safe concurrent manner.
    """

    def __init__(self, engine: Union[NeurolangDL, NeurolangPDL]) -> None:
        self.engines = set((engine,))
        self.counter = 1
        self.sema = BoundedSemaphore(value=1)

    def add_engine(self, engine: Union[NeurolangDL, NeurolangPDL]) -> None:
        """
        Add an engine to the set and increase the total number of engines
        in the set.

        Parameters
        ----------
        engine : Union[NeurolangDL, NeurolangPDL]
            the engine to add.
        """
        self.engines.add(engine)
        self.counter += 1
        self.sema = BoundedSemaphore(value=self.counter)

    @contextmanager
    def engine(self, timeout: int = None):
        """
        ContextManager to safely acquire an engine from the set.
        This method will block on the semaphore until an engine can be
        safely used from this set, and then yield it.
        At the end, the engine is put back in the set and the semaphore
        value is released.

        Parameters
        ----------
        timout : int
            When invoked with a timeout other than None, it will block for
            at most timeout seconds.

        Yields
        -------
        Union[NeurolangDL, NeurolangPDL]
            an engine.
        """
        lock = self.sema.acquire(timeout=timeout)
        if lock:
            try:
                engine = self.engines.pop()
                yield engine
            finally:
                self.engines.add(engine)
                self.sema.release()
        else:
            yield None


class NeurolangEngineConfiguration:
    """
    A NeurolangEngineConfiguration is a combination of an id key and
    a method which returns a Neurolang instance.
    """

    @abstractproperty
    def key(self):
        pass

    @abstractproperty
    def atlas(self):
        pass

    @abstractmethod
    def create(self) -> Union[NeurolangDL, NeurolangPDL]:
        pass

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NeurolangEngineConfiguration):
            return super().__eq__(other)
        return self.key == other.key

    def __hash__(self) -> int:
        return hash(self.key)


class NeurosynthEngineConf(NeurolangEngineConfiguration):
    def __init__(self, data_dir: Path, resolution=None) -> None:
        super().__init__()
        self.data_dir = data_dir
        self.resolution = resolution
        self._mni_atlas = None
        self._mni_brain_mask = None

    @property
    def key(self):
        return "neurosynth"

    @property
    def atlas(self):
        if self._mni_atlas is None:
            self._mni_atlas = nib.load(
                datasets.fetch_icbm152_2009(
                    data_dir=str(self.data_dir / "icbm")
                )["t1"]
            )
        return self._mni_atlas

    @property
    def brain_mask(self):
        if self._mni_brain_mask is None:
            self._mni_brain_mask = nib.load(
                datasets.fetch_icbm152_2009(
                    data_dir=str(self.data_dir / "icbm")
                )["mask"]
            )
        return self._mni_brain_mask

    def create(self) -> NeurolangPDL:
        mask = self.brain_mask
        if self.resolution is not None:
            mask = image.resample_img(mask, np.eye(3) * self.resolution)
        nl = init_frontend(mask)
        add_ploting_functions(nl)
        load_neurosynth_data(self.data_dir, nl, mask)
        return nl


def load_neurosynth_data(data_dir: Path, nl, mni_mask: nib.Nifti1Image):
    ns_database_fn, ns_features_fn = datasets.utils._fetch_files(
        data_dir / "neurosynth",
        [
            (
                "database.txt",
                "https://github.com/neurosynth/neurosynth-data/raw/e8f27c4a9a44dbfbc0750366166ad2ba34ac72d6/current_data.tar.gz",
                {"uncompress": True},
            ),
            (
                "features.txt",
                "https://github.com/neurosynth/neurosynth-data/raw/e8f27c4a9a44dbfbc0750366166ad2ba34ac72d6/current_data.tar.gz",
                {"uncompress": True},
            ),
        ],
    )

    activations = pd.read_csv(ns_database_fn, sep="\t")
    activations["id"] = activations["id"].apply(StudyID)
    mni_peaks = activations.loc[activations.space == "MNI"][
        ["x", "y", "z", "id"]
    ]
    non_mni_peaks = activations.loc[activations.space == "TAL"][
        ["x", "y", "z", "id"]
    ]
    proj_mat = np.linalg.pinv(
        np.array(
            [
                [0.9254, 0.0024, -0.0118, -1.0207],
                [-0.0048, 0.9316, -0.0871, -1.7667],
                [0.0152, 0.0883, 0.8924, 4.0926],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ).T
    )
    projected = np.round(
        np.dot(
            np.hstack(
                (
                    non_mni_peaks[["x", "y", "z"]].values,
                    np.ones((len(non_mni_peaks), 1)),
                )
            ),
            proj_mat,
        )[:, 0:3]
    )
    projected_df = pd.DataFrame(
        np.hstack([projected, non_mni_peaks[["id"]].values]),
        columns=["x", "y", "z", "id"],
    )
    peak_data = pd.concat([projected_df, mni_peaks]).astype(
        {"x": int, "y": int, "z": int}
    )
    study_ids = peak_data[["id"]].drop_duplicates()

    features = pd.read_csv(ns_features_fn, sep="\t")
    features.rename(columns={"pmid": "id"}, inplace=True)

    term_data = pd.melt(
        features,
        var_name="term",
        id_vars="id",
        value_name="tfidf",
    ).query("tfidf > 0")[["term", "tfidf", "id"]]
    term_data["id"] = term_data["id"].apply(StudyID)

    nl.add_tuple_set(peak_data, name="PeakReported")
    nl.add_tuple_set(study_ids, name="Study")
    nl.add_tuple_set(term_data, name="TermInStudyTFIDF")
    nl.add_uniform_probabilistic_choice_over_set(
        study_ids, name="SelectedStudy"
    )
    nl.add_tuple_set(
        np.round(
            nib.affines.apply_affine(
                mni_mask.affine,
                np.transpose(mni_mask.get_fdata().astype(int).nonzero()),
            )
        ).astype(int),
        name="Voxel",
    )


def init_frontend(mni_mask):
    """
    Create a Neurolang Probabilistic engine and add some aggregation methods.

    Returns
    -------
    NeurolangPDL
        the Neurolang engine
    """
    nl = NeurolangPDL()

    @nl.add_symbol
    def agg_create_region_ijk(
        i: Iterable, j: Iterable, k: Iterable
    ) -> ExplicitVBR:
        voxels = np.c_[i, j, k]
        return ExplicitVBR(voxels, mni_mask.affine, image_dim=mni_mask.shape)

    @nl.add_symbol
    def agg_create_region_overlay_ijk(
        i: Iterable, j: Iterable, k: Iterable, p: Iterable
    ) -> ExplicitVBR:
        voxels = np.c_[i, j, k]
        return ExplicitVBROverlay(
            voxels, mni_mask.affine, p, image_dim=mni_mask.shape
        )

    @nl.add_symbol
    def agg_create_region(
        x: Iterable, y: Iterable, z: Iterable
    ) -> ExplicitVBR:
        mni_coords = np.c_[x, y, z]
        voxels = np.round(
            nib.affines.apply_affine(
                np.linalg.inv(mni_mask.affine),
                mni_coords.astype(float),
            )
        ).astype(int)
        return ExplicitVBR(voxels, mni_mask.affine, image_dim=mni_mask.shape)

    @nl.add_symbol
    def agg_create_region_overlay(
        x: Iterable, y: Iterable, z: Iterable, p: Iterable
    ) -> ExplicitVBR:
        mni_coords = np.c_[x, y, z]
        voxels = np.round(
            nib.affines.apply_affine(
                np.linalg.inv(mni_mask.affine),
                mni_coords.astype(float),
            )
        ).astype(int)
        return ExplicitVBROverlay(
            voxels, mni_mask.affine, p, image_dim=mni_mask.shape
        )

    @nl.add_symbol
    def startswith(prefix: str, s: str) -> bool:
        """Describe the prefix of string `s`.

        Parameters
        ----------
        prefix : str
            prefix to query.
        s : str
            string to check whether its
            prefixed by `s`.

        Returns
        -------
        bool
            whether `s` is prefixed by
            `prefix`.
        """
        return s.startswith(prefix)

    @nl.add_symbol
    def principal_direction(s: ExplicitVBR, direction: str, eps=1e-6) -> bool:
        """Describe the principal direction of
        the extension of a volumetric region.

        Parameters
        ----------
        s : ExplicitVBR
            region to analyse the principal
            direction of its extension.
        direction : str
            principal directions, one of
            `LR`, `AP`, `SI`, for the directions
            left-right, anterio-posterior, and
            superior inferior respectively.
        eps : float, optional
            minimum difference on between
            directional standard deviations,
            by default 1e-6.

        Returns
        -------
        bool
            wether the principal variance of
            `s` is `direction`.
        """
        # Assuming RAS coding os the xyz space.
        c = ["LR", "AP", "SI"]

        s_xyz = s.to_xyz()
        cov = np.cov(s_xyz.T)
        evals, evecs = np.linalg.eig(cov)
        i = np.argmax(np.abs(evals))
        abs_max_evec = np.abs(evecs[:, i].squeeze())
        sort_dir = np.argsort(abs_max_evec)
        if (
            np.abs(abs_max_evec[sort_dir[-1]] - abs_max_evec[sort_dir[-2]])
            < eps
        ):
            return False
        else:
            main_dir = c[sort_dir[-1]]
        return (direction == main_dir) or (direction[::-1] == main_dir)

    return nl


class DestrieuxEngineConf(NeurolangEngineConfiguration):
    def __init__(self, data_dir: Path, resolution: int = None) -> None:
        super().__init__()
        self.data_dir = data_dir
        self.resolution = resolution
        self._mni_atlas = None

    @property
    def key(self):
        return "destrieux"

    @property
    def atlas(self):
        if self._mni_atlas is None:
            self._mni_atlas = nib.load(
                datasets.fetch_icbm152_2009(
                    data_dir=str(self.data_dir / "icbm")
                )["t1"]
            )
        return self._mni_atlas

    def create(self) -> NeurolangPDL:
        mask = self.atlas
        if self.resolution is not None:
            mask = image.resample_img(mask, np.eye(3) * self.resolution)
        nl = init_frontend(mask)
        nl.add_symbol(
            region_union,
            name="region_union",
            type_=Callable[[Iterable[ExplicitVBR]], ExplicitVBR],
        )

        load_destrieux_atlas(self.data_dir, nl)
        return nl


def load_destrieux_atlas(data_dir, nl):
    destrieux_atlas = datasets.fetch_atlas_destrieux_2009(
        data_dir=str(data_dir / "destrieux")
    )

    destrieux_atlas_images = nib.load(destrieux_atlas["maps"])
    destrieux_atlas_labels = {
        label: str(name.decode("utf8").replace("-", " ").replace("_", " "))
        for label, name in destrieux_atlas["labels"]
        if name != b"Background"
    }
    nl.add_atlas_set(
        "destrieux", destrieux_atlas_labels, destrieux_atlas_images
    )


def add_ploting_functions(nl: Union[NeurolangDL, NeurolangPDL]):
    matplotlib.use('Agg')

    @nl.add_symbol
    def agg_kde(terms: Iterable, probs: Iterable) -> matplotlib.figure.Figure:
        """
        Create a kde plot showing prob distribution per term.

        Parameters
        ----------
        terms : Iterable[str]
            the terms
        probs: Iterable[float]
            the probs

        Returns
        -------
        matplotlib.figure.Figure
            a Figure
        """
        df = pd.DataFrame(
            {
                "terms": terms,
                "probs": probs,
            }
        )
        fig, ax = plt.subplots()
        fig.suptitle("Distribution of probs / term")
        sns.kdeplot(ax=ax, data=df[df.probs > 0.002], x="probs", hue="terms")
        return fig
